#include "pmom.h"
#include "anomalous.h"
#include "gamma1.h"
#include "gamma2.h"
#include "gamma4ns.h"
#include "gamma4ns_NF.h"
#include "gamma4sg.h"
#include "gamma4pqqps.h"
#include "gamma4pqgps.h"
#include "gamma4pgqps.h"
#include "gamma4pggps.h"
#include "mesq.h"
#include "pegasus.h"
#include "resconst.h"
#include "settings.h"
#include "accel.h"
#include "constants.h"
#include "psi.h"

#include <iostream>
#include <iomanip>

using namespace resconst;

complex <double> *pmom::gamma1qq;
complex <double> *pmom::gamma1qqqed;  // QED
complex <double> *pmom::gamma1qqb;
complex <double> *pmom::gamma1qqp;
complex <double> *pmom::gamma1qqbp;
complex <double> *pmom::gamma1qg;
complex <double> *pmom::gamma1qgamma; // QED
complex <double> *pmom::gamma1gq;
complex <double> *pmom::gamma1gg;

complex <double> *pmom::gamma2qq;
complex <double> *pmom::gamma2qqb;
complex <double> *pmom::gamma2qqp;
complex <double> *pmom::gamma2qqbp;
complex <double> *pmom::gamma2qg;
complex <double> *pmom::gamma2gq;
complex <double> *pmom::gamma2gg;

complex <double> *pmom::gamma3qq;
complex <double> *pmom::gamma3qqb;
complex <double> *pmom::gamma3qqp;
complex <double> *pmom::gamma3qqbp;
complex <double> *pmom::gamma3qg;
complex <double> *pmom::gamma3gq;
complex <double> *pmom::gamma3gg;

complex <double> *pmom::gamma4qq;
complex <double> *pmom::gamma4qqb;
complex <double> *pmom::gamma4qqp;
complex <double> *pmom::gamma4qqbp;
complex <double> *pmom::gamma4qg;
complex <double> *pmom::gamma4gq;
complex <double> *pmom::gamma4gg;

complex <double> *pmom::gamma1qq_1;
complex <double> *pmom::gamma1qqqed_1;  // QED
complex <double> *pmom::gamma1qqb_1;
complex <double> *pmom::gamma1qqp_1;
complex <double> *pmom::gamma1qqbp_1;
complex <double> *pmom::gamma1qg_1;
complex <double> *pmom::gamma1qgamma_1; // QED
complex <double> *pmom::gamma1gq_1;
complex <double> *pmom::gamma1gg_1;
complex <double> *pmom::gamma2qq_1;
complex <double> *pmom::gamma2qqb_1;
complex <double> *pmom::gamma2qqp_1;
complex <double> *pmom::gamma2qqbp_1;
complex <double> *pmom::gamma2qg_1;
complex <double> *pmom::gamma2gq_1;
complex <double> *pmom::gamma2gg_1;
complex <double> *pmom::gamma3qq_1;
complex <double> *pmom::gamma3qqb_1;
complex <double> *pmom::gamma3qqp_1;
complex <double> *pmom::gamma3qqbp_1;
complex <double> *pmom::gamma3qg_1;
complex <double> *pmom::gamma3gq_1;
complex <double> *pmom::gamma3gg_1;
complex <double> *pmom::gamma4qq_1;
complex <double> *pmom::gamma4qqb_1;
complex <double> *pmom::gamma4qqp_1;
complex <double> *pmom::gamma4qqbp_1;
complex <double> *pmom::gamma4qg_1;
complex <double> *pmom::gamma4gq_1;
complex <double> *pmom::gamma4gg_1;

complex <double> *pmom::gamma1qq_2;
complex <double> *pmom::gamma1qqqed_2;  // QED
complex <double> *pmom::gamma1qqb_2;
complex <double> *pmom::gamma1qqp_2;
complex <double> *pmom::gamma1qqbp_2;
complex <double> *pmom::gamma1qg_2;
complex <double> *pmom::gamma1qgamma_2; // QED
complex <double> *pmom::gamma1gq_2;
complex <double> *pmom::gamma1gg_2;
complex <double> *pmom::gamma2qq_2;
complex <double> *pmom::gamma2qqb_2;
complex <double> *pmom::gamma2qqp_2;
complex <double> *pmom::gamma2qqbp_2;
complex <double> *pmom::gamma2qg_2;
complex <double> *pmom::gamma2gq_2;
complex <double> *pmom::gamma2gg_2;
complex <double> *pmom::gamma3qq_2;
complex <double> *pmom::gamma3qqb_2;
complex <double> *pmom::gamma3qqp_2;
complex <double> *pmom::gamma3qqbp_2;
complex <double> *pmom::gamma3qg_2;
complex <double> *pmom::gamma3gq_2;
complex <double> *pmom::gamma3gg_2;
complex <double> *pmom::gamma4qq_2;
complex <double> *pmom::gamma4qqb_2;
complex <double> *pmom::gamma4qqp_2;
complex <double> *pmom::gamma4qqbp_2;
complex <double> *pmom::gamma4qg_2;
complex <double> *pmom::gamma4gq_2;
complex <double> *pmom::gamma4gg_2;

void pmom::allocate()
{
  if (opts.mellin1d)
    {
      gamma1qq = new complex <double> [mellinint::mdim*2];
      gamma1qqqed = new complex <double> [mellinint::mdim*2];  // QED
      gamma1qqb = new complex <double> [mellinint::mdim*2];
      gamma1qqp = new complex <double> [mellinint::mdim*2];
      gamma1qqbp = new complex <double> [mellinint::mdim*2];
      gamma1qg = new complex <double> [mellinint::mdim*2];
      gamma1qgamma = new complex <double> [mellinint::mdim*2]; // QED
      gamma1gq = new complex <double> [mellinint::mdim*2];
      gamma1gg = new complex <double> [mellinint::mdim*2];

      gamma2qq = new complex <double> [mellinint::mdim*2];
      gamma2qqb = new complex <double> [mellinint::mdim*2];
      gamma2qqp = new complex <double> [mellinint::mdim*2];
      gamma2qqbp = new complex <double> [mellinint::mdim*2];
      gamma2qg = new complex <double> [mellinint::mdim*2];
      gamma2gq = new complex <double> [mellinint::mdim*2];
      gamma2gg = new complex <double> [mellinint::mdim*2];

      gamma3qq = new complex <double> [mellinint::mdim*2];
      gamma3qqb = new complex <double> [mellinint::mdim*2];
      gamma3qqp = new complex <double> [mellinint::mdim*2];
      gamma3qqbp = new complex <double> [mellinint::mdim*2];
      gamma3qg = new complex <double> [mellinint::mdim*2];
      gamma3gq = new complex <double> [mellinint::mdim*2];
      gamma3gg = new complex <double> [mellinint::mdim*2];

      gamma4qq = new complex <double> [mellinint::mdim*2];
      gamma4qqb = new complex <double> [mellinint::mdim*2];
      gamma4qqp = new complex <double> [mellinint::mdim*2];
      gamma4qqbp = new complex <double> [mellinint::mdim*2];
      gamma4qg = new complex <double> [mellinint::mdim*2];
      gamma4gq = new complex <double> [mellinint::mdim*2];
      gamma4gg = new complex <double> [mellinint::mdim*2];
    }
  else
    {
      gamma1qq_1 = new complex <double> [mellinint::mdim*2];
      gamma1qqqed_1 = new complex <double> [mellinint::mdim*2];  // QED
      gamma1qqb_1 = new complex <double> [mellinint::mdim*2];
      gamma1qqp_1 = new complex <double> [mellinint::mdim*2];
      gamma1qqbp_1 = new complex <double> [mellinint::mdim*2];
      gamma1qg_1 = new complex <double> [mellinint::mdim*2];
      gamma1qgamma_1 = new complex <double> [mellinint::mdim*2];     // QED
      gamma1gq_1 = new complex <double> [mellinint::mdim*2];
      gamma1gg_1 = new complex <double> [mellinint::mdim*2];
      gamma2qq_1 = new complex <double> [mellinint::mdim*2];
      gamma2qqb_1 = new complex <double> [mellinint::mdim*2];
      gamma2qqp_1 = new complex <double> [mellinint::mdim*2];
      gamma2qqbp_1 = new complex <double> [mellinint::mdim*2];
      gamma2qg_1 = new complex <double> [mellinint::mdim*2];
      gamma2gq_1 = new complex <double> [mellinint::mdim*2];
      gamma2gg_1 = new complex <double> [mellinint::mdim*2];
      gamma3qq_1 = new complex <double> [mellinint::mdim*2];
      gamma3qqb_1 = new complex <double> [mellinint::mdim*2];
      gamma3qqp_1 = new complex <double> [mellinint::mdim*2];
      gamma3qqbp_1 = new complex <double> [mellinint::mdim*2];
      gamma3qg_1 = new complex <double> [mellinint::mdim*2];
      gamma3gq_1 = new complex <double> [mellinint::mdim*2];
      gamma3gg_1 = new complex <double> [mellinint::mdim*2];
      gamma4qq_1 = new complex <double> [mellinint::mdim*2];
      gamma4qqb_1 = new complex <double> [mellinint::mdim*2];
      gamma4qqp_1 = new complex <double> [mellinint::mdim*2];
      gamma4qqbp_1 = new complex <double> [mellinint::mdim*2];
      gamma4qg_1 = new complex <double> [mellinint::mdim*2];
      gamma4gq_1 = new complex <double> [mellinint::mdim*2];
      gamma4gg_1 = new complex <double> [mellinint::mdim*2];

      gamma1qq_2 = new complex <double> [mellinint::mdim*2];
      gamma1qqqed_2 = new complex <double> [mellinint::mdim*2];  // QED
      gamma1qqb_2 = new complex <double> [mellinint::mdim*2];
      gamma1qqp_2 = new complex <double> [mellinint::mdim*2];
      gamma1qqbp_2 = new complex <double> [mellinint::mdim*2];
      gamma1qg_2 = new complex <double> [mellinint::mdim*2];
      gamma1qgamma_2 = new complex <double> [mellinint::mdim*2]; // QED
      gamma1gq_2 = new complex <double> [mellinint::mdim*2];
      gamma1gg_2 = new complex <double> [mellinint::mdim*2];
      gamma2qq_2 = new complex <double> [mellinint::mdim*2];
      gamma2qqb_2 = new complex <double> [mellinint::mdim*2];
      gamma2qqp_2 = new complex <double> [mellinint::mdim*2];
      gamma2qqbp_2 = new complex <double> [mellinint::mdim*2];
      gamma2qg_2 = new complex <double> [mellinint::mdim*2];
      gamma2gq_2 = new complex <double> [mellinint::mdim*2];
      gamma2gg_2 = new complex <double> [mellinint::mdim*2];
      gamma3qq_2 = new complex <double> [mellinint::mdim*2];
      gamma3qqb_2 = new complex <double> [mellinint::mdim*2];
      gamma3qqp_2 = new complex <double> [mellinint::mdim*2];
      gamma3qqbp_2 = new complex <double> [mellinint::mdim*2];
      gamma3qg_2 = new complex <double> [mellinint::mdim*2];
      gamma3gq_2 = new complex <double> [mellinint::mdim*2];
      gamma3gg_2 = new complex <double> [mellinint::mdim*2];
      gamma4qq_2 = new complex <double> [mellinint::mdim*2];
      gamma4qqb_2 = new complex <double> [mellinint::mdim*2];
      gamma4qqp_2 = new complex <double> [mellinint::mdim*2];
      gamma4qqbp_2 = new complex <double> [mellinint::mdim*2];
      gamma4qg_2 = new complex <double> [mellinint::mdim*2];
      gamma4gq_2 = new complex <double> [mellinint::mdim*2];
      gamma4gg_2 = new complex <double> [mellinint::mdim*2];
    }
}

void pmom::calc()
{
  //Evaluate Mellin moments of the QCD splitting functions at LO, NLO, and NNLO
  complex <double> psg_qq,psg_qg,psg_gq,psg_gg,pns_p,pns_m,pns_v;
  complex <double> pqqS_p_pqqbS,pqqS_m_pqqbS,pqqS,pqqbS,pqqV,pqqbV;
  double norm;
  double nf = 5.;
  int nfl = 5-3;
  int sign = mesq::positive;
  if (opts.mellin1d)
    for (int i = 0; i < mellinint::mdim; i++)
      {
	//Flavour decompostion from https://arxiv.org/pdf/hep-ph/0408244.pdf Eqs. (2.14), (2.15)	

	//gamma1 (LO)
	psg_qq = cx(psg0_.p0sg_[0][0][nfl][i]);
	psg_qg = cx(psg0_.p0sg_[1][0][nfl][i]);
	psg_gq = cx(psg0_.p0sg_[0][1][nfl][i]);
	psg_gg = cx(psg0_.p0sg_[1][1][nfl][i]);
	pns_p = cx(pns0_.p0ns_[nfl][i]);
	pns_m = cx(pns0_.p0ns_[nfl][i]);
	pns_v = cx(pns0_.p0ns_[nfl][i]);

	// //check moments with Eqs. 21-34 of https://arxiv.org/pdf/2111.15561.pdf
	// cout << endl;
	// for (int nn = 1; nn <= 4; nn++)
	//   {
	//     complex <double> NNN  = 2.*nn;
	//     double NNFF = 4.;
	//     double PQQ0 = real(gamma1pqq(NNN,NNFF))/4./M_PI;
	//     double PQG0 = real(gamma1pqg(NNN,NNFF))/4./M_PI;
	//     double PGQ0 = real(gamma1pgq(NNN,NNFF))/4./M_PI;
	//     double PGG0 = real(gamma1pgg(NNN,NNFF))/4./M_PI;
	// 
	//     double PQQ1 = real(gamma2pqq(NNN,NNFF))/pow(4.*M_PI,2)/PQQ0;
	//     double PQG1 = real(gamma2pqg(NNN,NNFF))/pow(4.*M_PI,2)/PQG0;
	//     double PGQ1 = real(gamma2pgq(NNN,NNFF))/pow(4.*M_PI,2)/PGQ0;
	//     double PGG1 = real(gamma2pgg(NNN,NNFF))/pow(4.*M_PI,2)/PGG0;
	// 
	//     double rhoqqps = (opts.gam4SGqq > 0) ? -0.501+(0.0-(-0.501)) *opts.gam4SGqq : -0.501-(-0.7-(-0.501))*opts.gam4SGqq;
	//     double rhoqg   = (opts.gam4SGqg > 0) ? -1.754+(-0.9-(-1.754))*opts.gam4SGqg : -1.754-(-2.5-(-1.754))*opts.gam4SGqg;
	//     double rhogq   = (opts.gam4SGgq > 0) ? -1.784+(-1.5-(-1.784))*opts.gam4SGgq : -1.784-(-1.8-(-1.784))*opts.gam4SGgq;
	//     double rhogg   = (opts.gam4SGgg > 0) ? 19.245+(15-19.245)   *opts.gam4SGgg : 19.245-(-5-19.245)   *opts.gam4SGgg;
	//     fcomplex NN = fcx(NNN-1.);
	// 
	//     pns_p = -(gamma4nsp_A(NNN-1.,NNFF)*(1.+opts.gam4NSp)/2. + gamma4nsp_B(NNN-1.,NNFF)*(1.-opts.gam4NSp)/2.);
	//     pns_m = -(gamma4nsm_A(NNN-1.,NNFF)*(1.+opts.gam4NSm)/2. + gamma4nsm_B(NNN-1.,NNFF)*(1.-opts.gam4NSm)/2.);
	//     pns_v = -(gamma4nsv_A(NNN-1.,NNFF)*(1.+opts.gam4NSv)/2. + gamma4nsv_B(NNN-1.,NNFF)*(1.-opts.gam4NSv)/2.);
	// 
	//     psg_qq = NNFF*cx(pqqps3a_(NN,rhoqqps))+pns_p;
	//     psg_qg = NNFF*cx(pqg3a_(NN,rhoqg));
	//     psg_gq = cx(pgq3a_(NN,rhogq));
	//     psg_gg = cx(pgg3a_(NN,rhogg));
	//     double PQQ3 = real(psg_qq)/pow(4.*M_PI,4)/PQQ0;
	//     double PQG3 = real(psg_qg)/pow(4.*M_PI,4)/PQG0;
	//     double PGQ3 = real(psg_gq)/pow(4.*M_PI,4)/PGQ0;
	//     double PGG3 = real(psg_gg)/pow(4.*M_PI,4)/PGG0;
	// 
	//     //cout << setw(7) << "PQQ = " << setw(12) << PQQ0 << setw(6) << "as(1+" << setw(12) << PQQ1 << setw(6) << "as^2+" << setw(12) << PQQ3 << setw(6) << "as^3" << endl;
	//     //cout << setw(7) << "PQG = " << setw(12) << PQG0 << setw(6) << "as(1+" << setw(12) << PQG1 << setw(6) << "as^2+" << setw(12) << PQG3 << setw(6) << "as^3" << endl;
	//     //cout << setw(7) << "PGQ = " << setw(12) << PGQ0 << setw(6) << "as(1+" << setw(12) << PGQ1 << setw(6) << "as^2+" << setw(12) << PGQ3 << setw(6) << "as^3" << endl;
	//     //cout << setw(7) << "PGG = " << setw(12) << PGG0 << setw(6) << "as(1+" << setw(12) << PGG1 << setw(6) << "as^2+" << setw(12) << PGG3 << setw(6) << "as^3" << endl;
	//   }
	
	norm = 4.;
	pqqS_p_pqqbS = (psg_qq-pns_p)/nf/norm;
	pqqS_m_pqqbS = (pns_v-pns_m)/nf/norm;
	pqqS         = (pqqS_p_pqqbS+pqqS_m_pqqbS)/2.;
	pqqbS        = (pqqS_p_pqqbS-pqqS_m_pqqbS)/2.;
	pqqV         = (pns_p+pns_m)/2./norm;
	pqqbV        = (pns_p-pns_m)/2./norm;

	gamma1qq[index(i,sign)]   = pqqV +pqqS;
	gamma1qqqed[index(i,sign)]   = (pqqV +pqqS)*(3./4.);   // QED
	gamma1qqb[index(i,sign)]  = 0.;//pqqbV+pqqbS;
	gamma1qqp[index(i,sign)]  = 0.;//pqqS;
	gamma1qqbp[index(i,sign)] = 0.;//pqqbS;
	gamma1qg[index(i,sign)]   = psg_qg/2./nf/norm;
	gamma1qgamma[index(i,sign)]   = (psg_qg/2./nf/norm)*6.;    // QED
	gamma1gq[index(i,sign)]   = psg_gq/norm;
	gamma1gg[index(i,sign)]   = psg_gg/norm;

	//gamma2 (NLO)
	psg_qq = cx(psg1_.p1sg_[0][0][nfl][i]);
	psg_qg = cx(psg1_.p1sg_[1][0][nfl][i]);
	psg_gq = cx(psg1_.p1sg_[0][1][nfl][i]);
	psg_gg = cx(psg1_.p1sg_[1][1][nfl][i]);
	pns_p = cx(pns1_.p1ns_[0][nfl][i]);
	pns_m = cx(pns1_.p1ns_[1][nfl][i]);
	pns_v = cx(pns1_.p1ns_[2][nfl][i]);
	
	norm = pow(4.,2);
	pqqS_p_pqqbS = (psg_qq-pns_p)/nf/norm;
	pqqS_m_pqqbS = (pns_v-pns_m)/nf/norm;
	pqqS         = (pqqS_p_pqqbS+pqqS_m_pqqbS)/2.;
	pqqbS        = (pqqS_p_pqqbS-pqqS_m_pqqbS)/2.;
	pqqV         = (pns_p+pns_m)/2./norm;
	pqqbV        = (pns_p-pns_m)/2./norm;

	gamma2qq[index(i,sign)]   = pqqV +pqqS;
	gamma2qqb[index(i,sign)]  = pqqbV+pqqbS;
	gamma2qqp[index(i,sign)]  = pqqS;
	gamma2qqbp[index(i,sign)] = pqqbS;
	gamma2qg[index(i,sign)]   = psg_qg/2./nf/norm;
	gamma2gq[index(i,sign)]   = psg_gq/norm;
	gamma2gg[index(i,sign)]   = psg_gg/norm;
	  
	//gamma3 (NNLO)
	psg_qq = cx(psg2_.p2sg_[0][0][nfl][i]);
	psg_qg = cx(psg2_.p2sg_[1][0][nfl][i]);
	psg_gq = cx(psg2_.p2sg_[0][1][nfl][i]);
	psg_gg = cx(psg2_.p2sg_[1][1][nfl][i]);
	pns_p = cx(pns2_.p2ns_[0][nfl][i]);
	pns_m = cx(pns2_.p2ns_[1][nfl][i]);
	pns_v = cx(pns2_.p2ns_[2][nfl][i]);
	
	norm = pow(4.,3);
	pqqS_p_pqqbS = (psg_qq-pns_p)/nf/norm;
	pqqS_m_pqqbS = (pns_v-pns_m)/nf/norm;
	pqqS         = (pqqS_p_pqqbS+pqqS_m_pqqbS)/2.;
	pqqbS        = (pqqS_p_pqqbS-pqqS_m_pqqbS)/2.;
	pqqV         = (pns_p+pns_m)/2./norm;
	pqqbV        = (pns_p-pns_m)/2./norm;

	gamma3qq[index(i,sign)]   = pqqV +pqqS;
	gamma3qqb[index(i,sign)]  = pqqbV+pqqbS;
	gamma3qqp[index(i,sign)]  = pqqS;
	gamma3qqbp[index(i,sign)] = pqqbS;
	gamma3qg[index(i,sign)]   = psg_qg/2./nf/norm;
	gamma3gq[index(i,sign)]   = psg_gq/norm;
	gamma3gg[index(i,sign)]   = psg_gg/norm;

	//gamma4 (NNNLO)
	pns_p = -(gamma4nsp_A(mellinint::Np[i]-1.)*(1.+opts.gam4NSp)/2. + gamma4nsp_B(mellinint::Np[i]-1.)*(1.-opts.gam4NSp)/2.);
	pns_m = -(gamma4nsm_A(mellinint::Np[i]-1.)*(1.+opts.gam4NSm)/2. + gamma4nsm_B(mellinint::Np[i]-1.)*(1.-opts.gam4NSm)/2.);
	pns_v = -(gamma4nsv_A(mellinint::Np[i]-1.)*(1.+opts.gam4NSv)/2. + gamma4nsv_B(mellinint::Np[i]-1.)*(1.-opts.gam4NSv)/2.);

	if (opts.sgapprox == 0)
	  {
	    pns_p = -pns_p;
	    pns_m = -pns_m;
	    pns_v = -pns_v;
	  }
	
	//MSHT20 N3LO approximation
	if (opts.sgapprox == -1)
	  {
	    psg_qq = -(gamma4pqqps_A(mellinint::Np[i]-1.)*(1.+opts.gam4SGqq)/2. + gamma4pqqps_B(mellinint::Np[i]-1.)*(1.-opts.gam4SGqq)/2.) + pns_p; //minus sign to correct for sign convention in Eq.(4) of https://arxiv.org/pdf/2302.07593.pdf
	    psg_qg = -(gamma4pqgps_A(mellinint::Np[i]-1.)*(1.+opts.gam4SGqg)/2. + gamma4pqgps_B(mellinint::Np[i]-1.)*(1.-opts.gam4SGqg)/2.); //minus sign to correct for sign convention in Eq.(4) of https://arxiv.org/pdf/2302.07593.pdf
	    psg_gq = -(gamma4pgqps_A(mellinint::Np[i]-1.)*(1.+opts.gam4SGgq)/2. + gamma4pgqps_B(mellinint::Np[i]-1.)*(1.-opts.gam4SGgq)/2.); //minus sign to correct for sign convention in Eq.(4) of https://arxiv.org/pdf/2302.07593.pdf
	    psg_gg = -(gamma4pggps_A(mellinint::Np[i]-1.)*(1.+opts.gam4SGgg)/2. + gamma4pggps_B(mellinint::Np[i]-1.)*(1.-opts.gam4SGgg)/2.); //minus sign to correct for sign convention in Eq.(4) of https://arxiv.org/pdf/2302.07593.pdf
	    
	    //fcomplex NNN;
	    //cout << "pqq" << endl;
	    //double rhoqq1 = -0.7; double rhoqq2 = 0.0;
	    //NNN = fcx(2. -1.); cout << 2  << "  "  << 4.*cx(pqqps3a_(NNN,rhoqq1)) << "  " << -gamma4pqqps_A(2. -1.) << "  " << -gamma4pqqps_B(2. -1.) << "  "  << 4.*cx(pqqps3a_(NNN,rhoqq2))<< endl;
	    //NNN = fcx(4. -1.); cout << 4  << "  "  << 4.*cx(pqqps3a_(NNN,rhoqq1)) << "  " << -gamma4pqqps_A(4. -1.) << "  " << -gamma4pqqps_B(4. -1.) << "  "  << 4.*cx(pqqps3a_(NNN,rhoqq2))<< endl;
	    //NNN = fcx(6. -1.); cout << 6  << "  "  << 4.*cx(pqqps3a_(NNN,rhoqq1)) << "  " << -gamma4pqqps_A(6. -1.) << "  " << -gamma4pqqps_B(6. -1.) << "  "  << 4.*cx(pqqps3a_(NNN,rhoqq2))<< endl;
	    //NNN = fcx(8. -1.); cout << 8  << "  "  << 4.*cx(pqqps3a_(NNN,rhoqq1)) << "  " << -gamma4pqqps_A(8. -1.) << "  " << -gamma4pqqps_B(8. -1.) << "  "  << 4.*cx(pqqps3a_(NNN,rhoqq2))<< endl;
	    //NNN = fcx(10.-1.); cout << 10 << "  "  << 4.*cx(pqqps3a_(NNN,rhoqq1)) << "  " << -gamma4pqqps_A(10.-1.) << "  " << -gamma4pqqps_B(10.-1.) << "  "  << 4.*cx(pqqps3a_(NNN,rhoqq2))<< endl;
	    //NNN = fcx(12.-1.); cout << 12 << "  "  << 4.*cx(pqqps3a_(NNN,rhoqq1)) << "  " << -gamma4pqqps_A(12.-1.) << "  " << -gamma4pqqps_B(12.-1.) << "  "  << 4.*cx(pqqps3a_(NNN,rhoqq2))<< endl;
	    //NNN = fcx(14.-1.); cout << 14 << "  "  << 4.*cx(pqqps3a_(NNN,rhoqq1)) << "  " << -gamma4pqqps_A(14.-1.) << "  " << -gamma4pqqps_B(14.-1.) << "  "  << 4.*cx(pqqps3a_(NNN,rhoqq2))<< endl;
	    //NNN = fcx(16.-1.); cout << 16 << "  "  << 4.*cx(pqqps3a_(NNN,rhoqq1)) << "  " << -gamma4pqqps_A(16.-1.) << "  " << -gamma4pqqps_B(16.-1.) << "  "  << 4.*cx(pqqps3a_(NNN,rhoqq2))<< endl;
	    //NNN = fcx(18.-1.); cout << 18 << "  "  << 4.*cx(pqqps3a_(NNN,rhoqq1)) << "  " << -gamma4pqqps_A(18.-1.) << "  " << -gamma4pqqps_B(18.-1.) << "  "  << 4.*cx(pqqps3a_(NNN,rhoqq2))<< endl;
	    //NNN = fcx(20.-1.); cout << 20 << "  "  << 4.*cx(pqqps3a_(NNN,rhoqq1)) << "  " << -gamma4pqqps_A(20.-1.) << "  " << -gamma4pqqps_B(20.-1.) << "  "  << 4.*cx(pqqps3a_(NNN,rhoqq2))<< endl;
	    //
	    //cout << "pqg" << endl;
	    //double rhoqg1 = -0.9; double rhoqg2 = -2.5;
	    //NNN = fcx(2. -1.); cout << 2  << "  "  << 4.*cx(pqg3a_(NNN,rhoqg1)) << "  " << -gamma4pqgps_A(2. -1.) << "  " << -gamma4pqgps_B(2. -1.) << "  "  << 4.*cx(pqg3a_(NNN,rhoqg2))<< endl;
	    //NNN = fcx(4. -1.); cout << 4  << "  "  << 4.*cx(pqg3a_(NNN,rhoqg1)) << "  " << -gamma4pqgps_A(4. -1.) << "  " << -gamma4pqgps_B(4. -1.) << "  "  << 4.*cx(pqg3a_(NNN,rhoqg2))<< endl;
	    //NNN = fcx(6. -1.); cout << 6  << "  "  << 4.*cx(pqg3a_(NNN,rhoqg1)) << "  " << -gamma4pqgps_A(6. -1.) << "  " << -gamma4pqgps_B(6. -1.) << "  "  << 4.*cx(pqg3a_(NNN,rhoqg2))<< endl;
	    //NNN = fcx(8. -1.); cout << 8  << "  "  << 4.*cx(pqg3a_(NNN,rhoqg1)) << "  " << -gamma4pqgps_A(8. -1.) << "  " << -gamma4pqgps_B(8. -1.) << "  "  << 4.*cx(pqg3a_(NNN,rhoqg2))<< endl;
	    //NNN = fcx(10.-1.); cout << 10 << "  "  << 4.*cx(pqg3a_(NNN,rhoqg1)) << "  " << -gamma4pqgps_A(10.-1.) << "  " << -gamma4pqgps_B(10.-1.) << "  "  << 4.*cx(pqg3a_(NNN,rhoqg2))<< endl;
	    //NNN = fcx(12.-1.); cout << 12 << "  "  << 4.*cx(pqg3a_(NNN,rhoqg1)) << "  " << -gamma4pqgps_A(12.-1.) << "  " << -gamma4pqgps_B(12.-1.) << "  "  << 4.*cx(pqg3a_(NNN,rhoqg2))<< endl;
	    //NNN = fcx(14.-1.); cout << 14 << "  "  << 4.*cx(pqg3a_(NNN,rhoqg1)) << "  " << -gamma4pqgps_A(14.-1.) << "  " << -gamma4pqgps_B(14.-1.) << "  "  << 4.*cx(pqg3a_(NNN,rhoqg2))<< endl;
	    //NNN = fcx(16.-1.); cout << 16 << "  "  << 4.*cx(pqg3a_(NNN,rhoqg1)) << "  " << -gamma4pqgps_A(16.-1.) << "  " << -gamma4pqgps_B(16.-1.) << "  "  << 4.*cx(pqg3a_(NNN,rhoqg2))<< endl;
	    //NNN = fcx(18.-1.); cout << 18 << "  "  << 4.*cx(pqg3a_(NNN,rhoqg1)) << "  " << -gamma4pqgps_A(18.-1.) << "  " << -gamma4pqgps_B(18.-1.) << "  "  << 4.*cx(pqg3a_(NNN,rhoqg2))<< endl;
	    //NNN = fcx(20.-1.); cout << 20 << "  "  << 4.*cx(pqg3a_(NNN,rhoqg1)) << "  " << -gamma4pqgps_A(20.-1.) << "  " << -gamma4pqgps_B(20.-1.) << "  "  << 4.*cx(pqg3a_(NNN,rhoqg2))<< endl;
	    //NNN = fcx(22.-1.); cout << 22 << "  "  << 4.*cx(pqg3a_(NNN,rhoqg1)) << "  " << -gamma4pqgps_A(22.-1.) << "  " << -gamma4pqgps_B(22.-1.) << "  "  << 4.*cx(pqg3a_(NNN,rhoqg2))<< endl;
	    //
	    //cout << "pgq" << endl;
	    //double rhogq1 = -1.8; double rhogq2 = -1.5;
	    //NNN = fcx(2. -1.); cout << 2  << "  "  << cx(pgq3a_(NNN,rhogq1)) << "  " << -gamma4pgqps_A(2. -1.) << "  " << -gamma4pgqps_B(2. -1.) << "  "  << cx(pgq3a_(NNN,rhogq2))<< endl;
	    //NNN = fcx(4. -1.); cout << 4  << "  "  << cx(pgq3a_(NNN,rhogq1)) << "  " << -gamma4pgqps_A(4. -1.) << "  " << -gamma4pgqps_B(4. -1.) << "  "  << cx(pgq3a_(NNN,rhogq2))<< endl;
	    //NNN = fcx(6. -1.); cout << 6  << "  "  << cx(pgq3a_(NNN,rhogq1)) << "  " << -gamma4pgqps_A(6. -1.) << "  " << -gamma4pgqps_B(6. -1.) << "  "  << cx(pgq3a_(NNN,rhogq2))<< endl;
	    //NNN = fcx(8. -1.); cout << 8  << "  "  << cx(pgq3a_(NNN,rhogq1)) << "  " << -gamma4pgqps_A(8. -1.) << "  " << -gamma4pgqps_B(8. -1.) << "  "  << cx(pgq3a_(NNN,rhogq2))<< endl;
	    //NNN = fcx(10.-1.); cout << 10 << "  "  << cx(pgq3a_(NNN,rhogq1)) << "  " << -gamma4pgqps_A(10.-1.) << "  " << -gamma4pgqps_B(10.-1.) << "  "  << cx(pgq3a_(NNN,rhogq2))<< endl;
	    //NNN = fcx(12.-1.); cout << 12 << "  "  << cx(pgq3a_(NNN,rhogq1)) << "  " << -gamma4pgqps_A(12.-1.) << "  " << -gamma4pgqps_B(12.-1.) << "  "  << cx(pgq3a_(NNN,rhogq2))<< endl;
	    //NNN = fcx(14.-1.); cout << 14 << "  "  << cx(pgq3a_(NNN,rhogq1)) << "  " << -gamma4pgqps_A(14.-1.) << "  " << -gamma4pgqps_B(14.-1.) << "  "  << cx(pgq3a_(NNN,rhogq2))<< endl;
	    //NNN = fcx(16.-1.); cout << 16 << "  "  << cx(pgq3a_(NNN,rhogq1)) << "  " << -gamma4pgqps_A(16.-1.) << "  " << -gamma4pgqps_B(16.-1.) << "  "  << cx(pgq3a_(NNN,rhogq2))<< endl;
	    //NNN = fcx(18.-1.); cout << 18 << "  "  << cx(pgq3a_(NNN,rhogq1)) << "  " << -gamma4pgqps_A(18.-1.) << "  " << -gamma4pgqps_B(18.-1.) << "  "  << cx(pgq3a_(NNN,rhogq2))<< endl;
	    //NNN = fcx(20.-1.); cout << 20 << "  "  << cx(pgq3a_(NNN,rhogq1)) << "  " << -gamma4pgqps_A(20.-1.) << "  " << -gamma4pgqps_B(20.-1.) << "  "  << cx(pgq3a_(NNN,rhogq2))<< endl;
	    //
	    //cout << "pgg" << endl;
	    //double rhogg1 = -5; double rhogg2 = -15;
	    //NNN = fcx(2. -1.); cout << 2  << "  "  << cx(pgg3a_(NNN,rhogg1)) << "  " << -gamma4pggps_A(2. -1.) << "  " << -gamma4pggps_B(2. -1.) << "  "  << cx(pgg3a_(NNN,rhogg2))<< endl;
	    //NNN = fcx(4. -1.); cout << 4  << "  "  << cx(pgg3a_(NNN,rhogg1)) << "  " << -gamma4pggps_A(4. -1.) << "  " << -gamma4pggps_B(4. -1.) << "  "  << cx(pgg3a_(NNN,rhogg2))<< endl;
	    //NNN = fcx(6. -1.); cout << 6  << "  "  << cx(pgg3a_(NNN,rhogg1)) << "  " << -gamma4pggps_A(6. -1.) << "  " << -gamma4pggps_B(6. -1.) << "  "  << cx(pgg3a_(NNN,rhogg2))<< endl;
	    //NNN = fcx(8. -1.); cout << 8  << "  "  << cx(pgg3a_(NNN,rhogg1)) << "  " << -gamma4pggps_A(8. -1.) << "  " << -gamma4pggps_B(8. -1.) << "  "  << cx(pgg3a_(NNN,rhogg2))<< endl;
	    //NNN = fcx(10.-1.); cout << 10 << "  "  << cx(pgg3a_(NNN,rhogg1)) << "  " << -gamma4pggps_A(10.-1.) << "  " << -gamma4pggps_B(10.-1.) << "  "  << cx(pgg3a_(NNN,rhogg2))<< endl;
	    //NNN = fcx(12.-1.); cout << 12 << "  "  << cx(pgg3a_(NNN,rhogg1)) << "  " << -gamma4pggps_A(12.-1.) << "  " << -gamma4pggps_B(12.-1.) << "  "  << cx(pgg3a_(NNN,rhogg2))<< endl;
	    //NNN = fcx(14.-1.); cout << 14 << "  "  << cx(pgg3a_(NNN,rhogg1)) << "  " << -gamma4pggps_A(14.-1.) << "  " << -gamma4pggps_B(14.-1.) << "  "  << cx(pgg3a_(NNN,rhogg2))<< endl;
	    //NNN = fcx(16.-1.); cout << 16 << "  "  << cx(pgg3a_(NNN,rhogg1)) << "  " << -gamma4pggps_A(16.-1.) << "  " << -gamma4pggps_B(16.-1.) << "  "  << cx(pgg3a_(NNN,rhogg2))<< endl;
	    //NNN = fcx(18.-1.); cout << 18 << "  "  << cx(pgg3a_(NNN,rhogg1)) << "  " << -gamma4pggps_A(18.-1.) << "  " << -gamma4pggps_B(18.-1.) << "  "  << cx(pgg3a_(NNN,rhogg2))<< endl;
	    //NNN = fcx(20.-1.); cout << 20 << "  "  << cx(pgg3a_(NNN,rhogg1)) << "  " << -gamma4pggps_A(20.-1.) << "  " << -gamma4pggps_B(20.-1.) << "  "  << cx(pgg3a_(NNN,rhogg2))<< endl;
	    //cout << endl;
	  }
	if (opts.sgapprox == 0)
	  {
	    //Values from Table 8 of https://arxiv.org/pdf/2207.04739.pdf
	    double rhoqqps = -0.501; //-0.7 < aPS < 0.0  //-0.527; 
	    double rhoqg   = -1.754; //-2.5 < aQG < -0.9 //-1.666;
	    double rhogq   = -1.784; //-1.8 < aGQ < -1.5 //-1.795;
	    double rhogg   = 19.245; //-5 < aGG < 15     //20.345;
	    fcomplex NN = fcx(mellinint::Np[i]-1.);
	    psg_qq = cx(pqqps3a_bug_(NN,rhoqqps))*(opts.gam4SGqq+1.);
	    psg_qg = cx(pqg3a_(NN,rhoqg))    *(opts.gam4SGqg+1.);
	    psg_gq = cx(pgq3a_(NN,rhogq))    *(opts.gam4SGgq+1.);
	    psg_gg = cx(pgg3a_(NN,rhogg))    *(opts.gam4SGgg+1.);
	  }
	if (opts.sgapprox == 1)
	  {
	    //Values from Table 8 of https://arxiv.org/pdf/2207.04739.pdf
	    double rhoqqps = (opts.gam4SGqq > 0) ? -0.501+(0.0-(-0.501)) *opts.gam4SGqq : -0.501-(-0.7-(-0.501))*opts.gam4SGqq;
	    double rhoqg   = (opts.gam4SGqg > 0) ? -1.754+(-0.9-(-1.754))*opts.gam4SGqg : -1.754-(-2.5-(-1.754))*opts.gam4SGqg;
	    double rhogq   = (opts.gam4SGgq > 0) ? -1.784+(-1.5-(-1.784))*opts.gam4SGgq : -1.784-(-1.8-(-1.784))*opts.gam4SGgq;
	    double rhogg   = (opts.gam4SGgg > 0) ? 19.245+(15-19.245)   *opts.gam4SGgg : 19.245-(-5-19.245)   *opts.gam4SGgg;
	    fcomplex NN = fcx(mellinint::Np[i]-1.);
	    psg_qq = nf*cx(pqqps3a_(NN,rhoqqps))+pns_p;
	    psg_qg = nf*cx(pqg3a_(NN,rhoqg))    ; //nf factor
	    psg_gq = cx(pgq3a_(NN,rhogq))    ;
	    psg_gg = cx(pgg3a_(NN,rhogg))    ;
	  }
	//Levin transform from https://arxiv.org/pdf/1307.1843.pdf
	if (opts.sgapprox == 2)
	  {
	    psg_qq = accel::levin(cx(psg0_.p0sg_[0][0][nfl][i]),cx(psg1_.p1sg_[0][0][nfl][i]),cx(psg2_.p2sg_[0][0][nfl][i]))*(opts.gam4SGqq+1.);
	    psg_qg = accel::levin(cx(psg0_.p0sg_[1][0][nfl][i]),cx(psg1_.p1sg_[1][0][nfl][i]),cx(psg2_.p2sg_[1][0][nfl][i]))*(opts.gam4SGqg+1.);
	    psg_gq = accel::levin(cx(psg0_.p0sg_[0][1][nfl][i]),cx(psg1_.p1sg_[0][1][nfl][i]),cx(psg2_.p2sg_[0][1][nfl][i]))*(opts.gam4SGgq+1.);
	    psg_gg = accel::levin(cx(psg0_.p0sg_[1][1][nfl][i]),cx(psg1_.p1sg_[1][1][nfl][i]),cx(psg2_.p2sg_[1][1][nfl][i]))*(opts.gam4SGgg+1.);
	  }
	//Weninger delta transform from https://arxiv.org/pdf/1307.1843.pdf
	else if (opts.sgapprox == 3)
	  {
	    psg_qq = accel::weninger(cx(psg0_.p0sg_[0][0][nfl][i]),cx(psg1_.p1sg_[0][0][nfl][i]),cx(psg2_.p2sg_[0][0][nfl][i]))*(opts.gam4SGqq+1.);
	    psg_qg = accel::weninger(cx(psg0_.p0sg_[1][0][nfl][i]),cx(psg1_.p1sg_[1][0][nfl][i]),cx(psg2_.p2sg_[1][0][nfl][i]))*(opts.gam4SGqg+1.);
	    psg_gq = accel::weninger(cx(psg0_.p0sg_[0][1][nfl][i]),cx(psg1_.p1sg_[0][1][nfl][i]),cx(psg2_.p2sg_[0][1][nfl][i]))*(opts.gam4SGgq+1.);
	    psg_gg = accel::weninger(cx(psg0_.p0sg_[1][1][nfl][i]),cx(psg1_.p1sg_[1][1][nfl][i]),cx(psg2_.p2sg_[1][1][nfl][i]))*(opts.gam4SGgg+1.);
	  }
	//Levin transform on the series gamma/gamma1
	else if (opts.sgapprox == 4)
	  {
	    psg_qq = accel::levin(cx(psg1_.p1sg_[0][0][nfl][i])/cx(psg0_.p0sg_[0][0][nfl][i]),cx(psg2_.p2sg_[0][0][nfl][i])/cx(psg0_.p0sg_[0][0][nfl][i]))*cx(psg0_.p0sg_[0][0][nfl][i])*(opts.gam4SGqq+1.);
	    psg_qg = accel::levin(cx(psg1_.p1sg_[1][0][nfl][i])/cx(psg0_.p0sg_[1][0][nfl][i]),cx(psg2_.p2sg_[1][0][nfl][i])/cx(psg0_.p0sg_[1][0][nfl][i]))*cx(psg0_.p0sg_[1][0][nfl][i])*(opts.gam4SGqg+1.);
	    psg_gq = accel::levin(cx(psg1_.p1sg_[0][1][nfl][i])/cx(psg0_.p0sg_[0][1][nfl][i]),cx(psg2_.p2sg_[0][1][nfl][i])/cx(psg0_.p0sg_[0][1][nfl][i]))*cx(psg0_.p0sg_[0][1][nfl][i])*(opts.gam4SGgq+1.);
	    psg_gg = accel::levin(cx(psg1_.p1sg_[1][1][nfl][i])/cx(psg0_.p0sg_[1][1][nfl][i]),cx(psg2_.p2sg_[1][1][nfl][i])/cx(psg0_.p0sg_[1][1][nfl][i]))*cx(psg0_.p0sg_[1][1][nfl][i])*(opts.gam4SGgg+1.);
	  }
	//Take gamma4 = gamma3
	//else if (opts.sgapprox == 5)
	//  {
	//    psg_qq = cx(psg2_.p2sg_[0][0][nfl][i])*4.*M_PI*(opts.gam4SGqq+1.);
	//    psg_qg = cx(psg2_.p2sg_[1][0][nfl][i])*4.*M_PI*(opts.gam4SGqg+1.);
	//    psg_gq = cx(psg2_.p2sg_[0][1][nfl][i])*4.*M_PI*(opts.gam4SGgq+1.);
	//    psg_gg = cx(psg2_.p2sg_[1][1][nfl][i])*4.*M_PI*(opts.gam4SGgg+1.);
	//  }
	//cout << "  " << psg_gg << endl;
	
	norm = pow(4.,4);
	pqqS_p_pqqbS = (psg_qq-pns_p)/nf/norm;
	pqqS_m_pqqbS = (pns_v-pns_m)/nf/norm;
	pqqS         = (pqqS_p_pqqbS+pqqS_m_pqqbS)/2.;
	pqqbS        = (pqqS_p_pqqbS-pqqS_m_pqqbS)/2.;
	pqqV         = (pns_p+pns_m)/2./norm;
	pqqbV        = (pns_p-pns_m)/2./norm;

	gamma4qq[index(i,sign)]   = pqqV +pqqS;
	gamma4qqb[index(i,sign)]  = pqqbV+pqqbS;
	gamma4qqp[index(i,sign)]  = pqqS;
	gamma4qqbp[index(i,sign)] = pqqbS;
	gamma4qg[index(i,sign)]   = psg_qg/2./nf/norm;
	gamma4gq[index(i,sign)]   = psg_gq/norm;
	gamma4gg[index(i,sign)]   = psg_gg/norm;
      }
  else
    for (int i = 0; i < 2*mellinint::mdim; i++)
      {
	//Flavour decompostion from https://arxiv.org/pdf/hep-ph/0408244.pdf Eqs. (2.14), (2.15)	

	//gamma1 (LO)
	psg_qq = cx(psg0_.p0sg_[0][0][nfl][i]);
	psg_qg = cx(psg0_.p0sg_[1][0][nfl][i]);
	psg_gq = cx(psg0_.p0sg_[0][1][nfl][i]);
	psg_gg = cx(psg0_.p0sg_[1][1][nfl][i]);
	pns_p = cx(pns0_.p0ns_[nfl][i]);
	pns_m = cx(pns0_.p0ns_[nfl][i]);
	pns_v = cx(pns0_.p0ns_[nfl][i]);

	norm = 4.;
	pqqS_p_pqqbS = (psg_qq-pns_p)/nf/norm;
	pqqS_m_pqqbS = (pns_v-pns_m)/nf/norm;
	pqqS         = (pqqS_p_pqqbS+pqqS_m_pqqbS)/2.;
	pqqbS        = (pqqS_p_pqqbS-pqqS_m_pqqbS)/2.;
	pqqV         = (pns_p+pns_m)/2./norm;
	pqqbV        = (pns_p-pns_m)/2./norm;

	if (i < mellinint::mdim)
	  {
	    gamma1qq_1[index(i,sign)]   = pqqV +pqqS;
	    gamma1qqqed_1[index(i,sign)]   = (pqqV +pqqS)*(3./4.);     // QED
	    gamma1qqb_1[index(i,sign)]  = 0.;//pqqbV+pqqbS;
	    gamma1qqp_1[index(i,sign)]  = 0.;//pqqS;
	    gamma1qqbp_1[index(i,sign)] = 0.;//pqqbS;
	    gamma1qg_1[index(i,sign)]   = psg_qg/2./nf/norm;
	    gamma1qgamma_1[index(i,sign)]   = (psg_qg/2./nf/norm)*6.;  // QED
	    gamma1gq_1[index(i,sign)]   = psg_gq/norm;
	    gamma1gg_1[index(i,sign)]   = psg_gg/norm;
	    //cout << i << " Np1 " << mellinint::Np_1[i] << "  " << gamma1qg_1[index(i,sign)] << endl;
	  }
	else
	  {
	    gamma1qq_2[index(i-mellinint::mdim,sign)]   = pqqV +pqqS;
	    gamma1qqqed_2[index(i-mellinint::mdim,sign)]   = (pqqV +pqqS)*(3./4.);   // QED
	    gamma1qqb_2[index(i-mellinint::mdim,sign)]  = 0.;//pqqbV+pqqbS;
	    gamma1qqp_2[index(i-mellinint::mdim,sign)]  = 0.;//pqqS;
	    gamma1qqbp_2[index(i-mellinint::mdim,sign)] = 0.;//pqqbS;
	    gamma1qg_2[index(i-mellinint::mdim,sign)]   = psg_qg/2./nf/norm;
	    gamma1qgamma_2[index(i-mellinint::mdim,sign)]   = psg_qg/2./nf/norm*6.;   // QED
	    gamma1gq_2[index(i-mellinint::mdim,sign)]   = psg_gq/norm;
	    gamma1gg_2[index(i-mellinint::mdim,sign)]   = psg_gg/norm;
	    //cout << i-mellinint::mdim << " Np2  " << mellinint::Np_2[i-mellinint::mdim] << "  " << gamma1qg_2[index(i-mellinint::mdim,sign)] << endl;
	  }

	//gamma2 (NLO)
	psg_qq = cx(psg1_.p1sg_[0][0][nfl][i]);
	psg_qg = cx(psg1_.p1sg_[1][0][nfl][i]);
	psg_gq = cx(psg1_.p1sg_[0][1][nfl][i]);
	psg_gg = cx(psg1_.p1sg_[1][1][nfl][i]);
	pns_p = cx(pns1_.p1ns_[0][nfl][i]);
	pns_m = cx(pns1_.p1ns_[1][nfl][i]);
	pns_v = cx(pns1_.p1ns_[2][nfl][i]);

	norm = pow(4.,2);
	pqqS_p_pqqbS = (psg_qq-pns_p)/nf/norm;
	pqqS_m_pqqbS = (pns_v-pns_m)/nf/norm;
	pqqS         = (pqqS_p_pqqbS+pqqS_m_pqqbS)/2.;
	pqqbS        = (pqqS_p_pqqbS-pqqS_m_pqqbS)/2.;
	pqqV         = (pns_p+pns_m)/2./norm;
	pqqbV        = (pns_p-pns_m)/2./norm;

	if (i < mellinint::mdim)
	  {
	    gamma2qq_1[index(i,sign)]   = pqqV +pqqS;
	    gamma2qqb_1[index(i,sign)]  = pqqbV+pqqbS;
	    gamma2qqp_1[index(i,sign)]  = pqqS;
	    gamma2qqbp_1[index(i,sign)] = pqqbS;
	    gamma2qg_1[index(i,sign)]   = psg_qg/2./nf/norm;
	    gamma2gq_1[index(i,sign)]   = psg_gq/norm;
	    gamma2gg_1[index(i,sign)]   = psg_gg/norm;
	  }
	else
	  {
	    gamma2qq_2[index(i-mellinint::mdim,sign)]   = pqqV +pqqS;
	    gamma2qqb_2[index(i-mellinint::mdim,sign)]  = pqqbV+pqqbS;
	    gamma2qqp_2[index(i-mellinint::mdim,sign)]  = pqqS;
	    gamma2qqbp_2[index(i-mellinint::mdim,sign)] = pqqbS;
	    gamma2qg_2[index(i-mellinint::mdim,sign)]   = psg_qg/2./nf/norm;
	    gamma2gq_2[index(i-mellinint::mdim,sign)]   = psg_gq/norm;
	    gamma2gg_2[index(i-mellinint::mdim,sign)]   = psg_gg/norm;
	  }
	
	//gamma3 (NNLO)
	psg_qq = cx(psg2_.p2sg_[0][0][nfl][i]);
	psg_qg = cx(psg2_.p2sg_[1][0][nfl][i]);
	psg_gq = cx(psg2_.p2sg_[0][1][nfl][i]);
	psg_gg = cx(psg2_.p2sg_[1][1][nfl][i]);
	pns_p = cx(pns2_.p2ns_[0][nfl][i]);
	pns_m = cx(pns2_.p2ns_[1][nfl][i]);
	pns_v = cx(pns2_.p2ns_[2][nfl][i]);
	
	norm = pow(4.,3);
	pqqS_p_pqqbS = (psg_qq-pns_p)/nf/norm;
	pqqS_m_pqqbS = (pns_v-pns_m)/nf/norm;
	pqqS         = (pqqS_p_pqqbS+pqqS_m_pqqbS)/2.;
	pqqbS        = (pqqS_p_pqqbS-pqqS_m_pqqbS)/2.;
	pqqV         = (pns_p+pns_m)/2./norm;
	pqqbV        = (pns_p-pns_m)/2./norm;

	if (i < mellinint::mdim)
	  {
	    gamma3qq_1[index(i,sign)]   = pqqV +pqqS;
	    gamma3qqb_1[index(i,sign)]  = pqqbV+pqqbS;
	    gamma3qqp_1[index(i,sign)]  = pqqS;
	    gamma3qqbp_1[index(i,sign)] = pqqbS;
	    gamma3qg_1[index(i,sign)]   = psg_qg/2./nf/norm;
	    gamma3gq_1[index(i,sign)]   = psg_gq/norm;
	    gamma3gg_1[index(i,sign)]   = psg_gg/norm;
	  }
	else
	  {
	    gamma3qq_2[index(i-mellinint::mdim,sign)]   = pqqV +pqqS;
	    gamma3qqb_2[index(i-mellinint::mdim,sign)]  = pqqbV+pqqbS;
	    gamma3qqp_2[index(i-mellinint::mdim,sign)]  = pqqS;
	    gamma3qqbp_2[index(i-mellinint::mdim,sign)] = pqqbS;
	    gamma3qg_2[index(i-mellinint::mdim,sign)]   = psg_qg/2./nf/norm;
	    gamma3gq_2[index(i-mellinint::mdim,sign)]   = psg_gq/norm;
	    gamma3gg_2[index(i-mellinint::mdim,sign)]   = psg_gg/norm;
	  }

	//gamma4 (NNNLO)
	complex <double> N;
	if (i < mellinint::mdim)
	  N = mellinint::Np_1[i]-1.;
	else
	  N = mellinint::Np_2[i-mellinint::mdim]-1.;

	pns_p = -(gamma4nsp_A(N)*(1.+opts.gam4NSp)/2. + gamma4nsp_B(N)*(1.-opts.gam4NSp)/2.);
	pns_m = -(gamma4nsm_A(N)*(1.+opts.gam4NSm)/2. + gamma4nsm_B(N)*(1.-opts.gam4NSm)/2.);
	pns_v = -(gamma4nsv_A(N)*(1.+opts.gam4NSv)/2. + gamma4nsv_B(N)*(1.-opts.gam4NSv)/2.);

	if (opts.sgapprox == 0)
	  {
	    pns_p = -pns_p;
	    pns_m = -pns_m;
	    pns_v = -pns_v;
	  }
	
	//MSHT20 N3LO approximation
	if (opts.sgapprox == -1)
	  {
	    psg_qq = -(gamma4pqqps_A(N)*(1.+opts.gam4SGqq)/2. + gamma4pqqps_B(N)*(1.-opts.gam4SGqq)/2.) + pns_p; //minus sign to correct for sign convention in Eq.(4) of https://arxiv.org/pdf/2302.07593.pdf
	    psg_qg = -(gamma4pqgps_A(N)*(1.+opts.gam4SGqg)/2. + gamma4pqgps_B(N)*(1.-opts.gam4SGqg)/2.); //minus sign to correct for sign convention in Eq.(4) of https://arxiv.org/pdf/2302.07593.pdf
	    psg_gq = -(gamma4pgqps_A(N)*(1.+opts.gam4SGgq)/2. + gamma4pgqps_B(N)*(1.-opts.gam4SGgq)/2.); //minus sign to correct for sign convention in Eq.(4) of https://arxiv.org/pdf/2302.07593.pdf
	    psg_gg = -(gamma4pggps_A(N)*(1.+opts.gam4SGgg)/2. + gamma4pggps_B(N)*(1.-opts.gam4SGgg)/2.); //minus sign to correct for sign convention in Eq.(4) of https://arxiv.org/pdf/2302.07593.pdf
	  }
	if (opts.sgapprox == 0)
	  {
	    //Values from Table 8 of https://arxiv.org/pdf/2207.04739.pdf
	    double rhoqqps = -0.501; //-0.7 < aPS < 0.0  //-0.527; 
	    double rhoqg   = -1.754; //-2.5 < aQG < -0.9 //-1.666;
	    double rhogq   = -1.784; //-1.8 < aGQ < -1.5 //-1.795;
	    double rhogg   = 19.245; //-5 < aGG < 15     //20.345;
	    fcomplex NN = fcx(N);
	    psg_qq = cx(pqqps3a_bug_(NN,rhoqqps))*(opts.gam4SGqq+1.);
	    psg_qg = cx(pqg3a_(NN,rhoqg))    *(opts.gam4SGqg+1.);
	    psg_gq = cx(pgq3a_(NN,rhogq))    *(opts.gam4SGgq+1.);
	    psg_gg = cx(pgg3a_(NN,rhogg))    *(opts.gam4SGgg+1.);
	  }
	if (opts.sgapprox == 1)
	  {
	    //Values from Table 13 of https://arxiv.org/pdf/2207.04739.pdf
	    double rhoqqps = (opts.gam4SGqq > 0) ? -0.501+(0.0-(-0.501)) *opts.gam4SGqq : -0.501-(-0.7-(-0.501))*opts.gam4SGqq;
	    double rhoqg   = (opts.gam4SGqg > 0) ? -1.754+(-0.9-(-1.754))*opts.gam4SGqg : -1.754-(-2.5-(-1.754))*opts.gam4SGqg;
	    double rhogq   = (opts.gam4SGgq > 0) ? -1.784+(-1.5-(-1.784))*opts.gam4SGgq : -1.784-(-1.8-(-1.784))*opts.gam4SGgq;
	    double rhogg   = (opts.gam4SGgg > 0) ? 19.245+(15-19.245)   *opts.gam4SGgg : 19.245-(-5-19.245)   *opts.gam4SGgg;
	    fcomplex NN = fcx(N);
	    psg_qq = nf*cx(pqqps3a_(NN,rhoqqps))+pns_p;
	    psg_qg = nf*cx(pqg3a_(NN,rhoqg))    ;  //nf factor
	    psg_gq = cx(pgq3a_(NN,rhogq))    ;
	    psg_gg = cx(pgg3a_(NN,rhogg))    ;
	  }
	//Levin transform from https://arxiv.org/pdf/1307.1843.pdf
	if (opts.sgapprox == 2)
	  {
	    psg_qq = accel::levin(cx(psg0_.p0sg_[0][0][nfl][i]),cx(psg1_.p1sg_[0][0][nfl][i]),cx(psg2_.p2sg_[0][0][nfl][i]))*(opts.gam4SGqq+1.);
	    psg_qg = accel::levin(cx(psg0_.p0sg_[1][0][nfl][i]),cx(psg1_.p1sg_[1][0][nfl][i]),cx(psg2_.p2sg_[1][0][nfl][i]))*(opts.gam4SGqg+1.);
	    psg_gq = accel::levin(cx(psg0_.p0sg_[0][1][nfl][i]),cx(psg1_.p1sg_[0][1][nfl][i]),cx(psg2_.p2sg_[0][1][nfl][i]))*(opts.gam4SGgq+1.);
	    psg_gg = accel::levin(cx(psg0_.p0sg_[1][1][nfl][i]),cx(psg1_.p1sg_[1][1][nfl][i]),cx(psg2_.p2sg_[1][1][nfl][i]))*(opts.gam4SGgg+1.);
	  }
	//Weninger delta transform from https://arxiv.org/pdf/1307.1843.pdf
	else if (opts.sgapprox == 3)
	  {
	    psg_qq = accel::weninger(cx(psg0_.p0sg_[0][0][nfl][i]),cx(psg1_.p1sg_[0][0][nfl][i]),cx(psg2_.p2sg_[0][0][nfl][i]))*(opts.gam4SGqq+1.);
	    psg_qg = accel::weninger(cx(psg0_.p0sg_[1][0][nfl][i]),cx(psg1_.p1sg_[1][0][nfl][i]),cx(psg2_.p2sg_[1][0][nfl][i]))*(opts.gam4SGqg+1.);
	    psg_gq = accel::weninger(cx(psg0_.p0sg_[0][1][nfl][i]),cx(psg1_.p1sg_[0][1][nfl][i]),cx(psg2_.p2sg_[0][1][nfl][i]))*(opts.gam4SGgq+1.);
	    psg_gg = accel::weninger(cx(psg0_.p0sg_[1][1][nfl][i]),cx(psg1_.p1sg_[1][1][nfl][i]),cx(psg2_.p2sg_[1][1][nfl][i]))*(opts.gam4SGgg+1.);
	  }
	//Levin transform on the series gamma/gamma1
	else if (opts.sgapprox == 4)
	  {
	    psg_qq = accel::levin(cx(psg1_.p1sg_[0][0][nfl][i])/cx(psg0_.p0sg_[0][0][nfl][i]),cx(psg2_.p2sg_[0][0][nfl][i])/cx(psg0_.p0sg_[0][0][nfl][i]))*cx(psg0_.p0sg_[0][0][nfl][i])*(opts.gam4SGqq+1.);
	    psg_qg = accel::levin(cx(psg1_.p1sg_[1][0][nfl][i])/cx(psg0_.p0sg_[1][0][nfl][i]),cx(psg2_.p2sg_[1][0][nfl][i])/cx(psg0_.p0sg_[1][0][nfl][i]))*cx(psg0_.p0sg_[1][0][nfl][i])*(opts.gam4SGqg+1.);
	    psg_gq = accel::levin(cx(psg1_.p1sg_[0][1][nfl][i])/cx(psg0_.p0sg_[0][1][nfl][i]),cx(psg2_.p2sg_[0][1][nfl][i])/cx(psg0_.p0sg_[0][1][nfl][i]))*cx(psg0_.p0sg_[0][1][nfl][i])*(opts.gam4SGgq+1.);
	    psg_gg = accel::levin(cx(psg1_.p1sg_[1][1][nfl][i])/cx(psg0_.p0sg_[1][1][nfl][i]),cx(psg2_.p2sg_[1][1][nfl][i])/cx(psg0_.p0sg_[1][1][nfl][i]))*cx(psg0_.p0sg_[1][1][nfl][i])*(opts.gam4SGgg+1.);
	  }
	//Take gamma4 = gamma3
	//else if (opts.sgapprox == 5)
	//  {
	//    psg_qq = cx(psg2_.p2sg_[0][0][nfl][i])*4.*M_PI*(opts.gam4SGqq+1.);
	//    psg_qg = cx(psg2_.p2sg_[1][0][nfl][i])*4.*M_PI*(opts.gam4SGqg+1.);
	//    psg_gq = cx(psg2_.p2sg_[0][1][nfl][i])*4.*M_PI*(opts.gam4SGgq+1.);
	//    psg_gg = cx(psg2_.p2sg_[1][1][nfl][i])*4.*M_PI*(opts.gam4SGgg+1.);
	//  }

	norm = pow(4.,4);
	pqqS_p_pqqbS = (psg_qq-pns_p)/nf/norm;
	pqqS_m_pqqbS = (pns_v-pns_m)/nf/norm;
	pqqS         = (pqqS_p_pqqbS+pqqS_m_pqqbS)/2.;
	pqqbS        = (pqqS_p_pqqbS-pqqS_m_pqqbS)/2.;
	pqqV         = (pns_p+pns_m)/2./norm;
	pqqbV        = (pns_p-pns_m)/2./norm;

	if (i < mellinint::mdim)
	  {
	    gamma4qq_1[index(i,sign)]   = pqqV +pqqS;
	    gamma4qqb_1[index(i,sign)]  = pqqbV+pqqbS;
	    gamma4qqp_1[index(i,sign)]  = pqqS;
	    gamma4qqbp_1[index(i,sign)] = pqqbS;
	    gamma4qg_1[index(i,sign)]   = psg_qg/2./nf/norm;
	    gamma4gq_1[index(i,sign)]   = psg_gq/norm;
	    gamma4gg_1[index(i,sign)]   = psg_gg/norm;
	  }
	else
	  {
	    gamma4qq_2[index(i-mellinint::mdim,sign)]   = pqqV +pqqS;
	    gamma4qqb_2[index(i-mellinint::mdim,sign)]  = pqqbV+pqqbS;
	    gamma4qqp_2[index(i-mellinint::mdim,sign)]  = pqqS;
	    gamma4qqbp_2[index(i-mellinint::mdim,sign)] = pqqbS;
	    gamma4qg_2[index(i-mellinint::mdim,sign)]   = psg_qg/2./nf/norm;
	    gamma4gq_2[index(i-mellinint::mdim,sign)]   = psg_gq/norm;
	    gamma4gg_2[index(i-mellinint::mdim,sign)]   = psg_gg/norm;
	  }
      }

  if (opts.cmw)
    if (opts.mellin1d)
      for (int i = 0; i < mellinint::mdim; i++)
	{
	  complex <double> gamma2qqCMW   = gamma2qq[index(i,sign)]   - K1* gamma1qq[index(i,sign)]   ;
	  complex <double> gamma2qqbCMW  = gamma2qqb[index(i,sign)]  - K1* gamma1qqb[index(i,sign)]  ;
	  complex <double> gamma2qqpCMW  = gamma2qqp[index(i,sign)]  - K1* gamma1qqp[index(i,sign)]  ;
	  complex <double> gamma2qqbpCMW = gamma2qqbp[index(i,sign)] - K1* gamma1qqbp[index(i,sign)] ;
	  complex <double> gamma2qgCMW   = gamma2qg[index(i,sign)]   - K1* gamma1qg[index(i,sign)]   ;
	  complex <double> gamma2gqCMW   = gamma2gq[index(i,sign)]   - K1* gamma1gq[index(i,sign)]   ;
	  complex <double> gamma2ggCMW   = gamma2gg[index(i,sign)]   - K1* gamma1gg[index(i,sign)]   ;

	  complex <double> gamma3qqCMW   = gamma3qq[index(i,sign)]   - 2.*K1* gamma2qq[index(i,sign)]   + 2.*pow(K1,2)*gamma1qq[index(i,sign)]   - K2*gamma1qq[index(i,sign)]   ;
	  complex <double> gamma3qqbCMW  = gamma3qqb[index(i,sign)]  - 2.*K1* gamma2qqb[index(i,sign)]  + 2.*pow(K1,2)*gamma1qqb[index(i,sign)]  - K2*gamma1qqb[index(i,sign)]  ;
	  complex <double> gamma3qqpCMW  = gamma3qqp[index(i,sign)]  - 2.*K1* gamma2qqp[index(i,sign)]  + 2.*pow(K1,2)*gamma1qqp[index(i,sign)]  - K2*gamma1qqp[index(i,sign)]  ;
	  complex <double> gamma3qqbpCMW = gamma3qqbp[index(i,sign)] - 2.*K1* gamma2qqbp[index(i,sign)] + 2.*pow(K1,2)*gamma1qqbp[index(i,sign)] - K2*gamma1qqbp[index(i,sign)] ;
	  complex <double> gamma3qgCMW   = gamma3qg[index(i,sign)]   - 2.*K1* gamma2qg[index(i,sign)]   + 2.*pow(K1,2)*gamma1qg[index(i,sign)]   - K2*gamma1qg[index(i,sign)]   ;
	  complex <double> gamma3gqCMW   = gamma3gq[index(i,sign)]   - 2.*K1* gamma2gq[index(i,sign)]   + 2.*pow(K1,2)*gamma1gq[index(i,sign)]   - K2*gamma1gq[index(i,sign)]   ;
	  complex <double> gamma3ggCMW   = gamma3gg[index(i,sign)]   - 2.*K1* gamma2gg[index(i,sign)]   + 2.*pow(K1,2)*gamma1gg[index(i,sign)]   - K2*gamma1gg[index(i,sign)]   ;
	  
	  complex <double> gamma4qqCMW   = gamma4qq[index(i,sign)]   - 3.*K1* gamma3qq[index(i,sign)]   + (5.*pow(K1,2)-2.*K2)*gamma2qq[index(i,sign)]   +(-5.*pow(K1,3)+5.*K1*K2-K3)*gamma1qq[index(i,sign)]   ;
	  complex <double> gamma4qqbCMW  = gamma4qqb[index(i,sign)]  - 3.*K1* gamma3qqb[index(i,sign)]  + (5.*pow(K1,2)-2.*K2)*gamma2qqb[index(i,sign)]  +(-5.*pow(K1,3)+5.*K1*K2-K3)*gamma1qqb[index(i,sign)]  ;
	  complex <double> gamma4qqpCMW  = gamma4qqp[index(i,sign)]  - 3.*K1* gamma3qqp[index(i,sign)]  + (5.*pow(K1,2)-2.*K2)*gamma2qqp[index(i,sign)]  +(-5.*pow(K1,3)+5.*K1*K2-K3)*gamma1qqp[index(i,sign)]  ;
	  complex <double> gamma4qqbpCMW = gamma4qqbp[index(i,sign)] - 3.*K1* gamma3qqbp[index(i,sign)] + (5.*pow(K1,2)-2.*K2)*gamma2qqbp[index(i,sign)] +(-5.*pow(K1,3)+5.*K1*K2-K3)*gamma1qqbp[index(i,sign)] ;
	  complex <double> gamma4qgCMW   = gamma4qg[index(i,sign)]   - 3.*K1* gamma3qg[index(i,sign)]   + (5.*pow(K1,2)-2.*K2)*gamma2qg[index(i,sign)]   +(-5.*pow(K1,3)+5.*K1*K2-K3)*gamma1qg[index(i,sign)]   ;
	  complex <double> gamma4gqCMW   = gamma4gq[index(i,sign)]   - 3.*K1* gamma3gq[index(i,sign)]   + (5.*pow(K1,2)-2.*K2)*gamma2gq[index(i,sign)]   +(-5.*pow(K1,3)+5.*K1*K2-K3)*gamma1gq[index(i,sign)]   ;
	  complex <double> gamma4ggCMW   = gamma4gg[index(i,sign)]   - 3.*K1* gamma3gg[index(i,sign)]   + (5.*pow(K1,2)-2.*K2)*gamma2gg[index(i,sign)]   +(-5.*pow(K1,3)+5.*K1*K2-K3)*gamma1gg[index(i,sign)]   ;

	  //Shoulde keep gamma in msbar scheme, to match PDF evolution in msbr scheme? --> mismatch with alphas in the gamma terms of Hst
	  //--> should keep both an msbar version for the PDF evolution and a cmw version for Hst and commute evolution
	  gamma2qq[index(i,sign)]   = gamma2qqCMW   ;
	  gamma2qqb[index(i,sign)]  = gamma2qqbCMW  ;
	  gamma2qqp[index(i,sign)]  = gamma2qqpCMW  ;
	  gamma2qqbp[index(i,sign)] = gamma2qqbpCMW ;
	  gamma2qg[index(i,sign)]   = gamma2qgCMW   ;
	  gamma2gq[index(i,sign)]   = gamma2gqCMW   ;
	  gamma2gg[index(i,sign)]   = gamma2ggCMW   ;
	  
	  gamma3qq[index(i,sign)]   = gamma3qqCMW   ;
	  gamma3qqb[index(i,sign)]  = gamma3qqbCMW  ;
	  gamma3qqp[index(i,sign)]  = gamma3qqpCMW  ;
	  gamma3qqbp[index(i,sign)] = gamma3qqbpCMW ;
	  gamma3qg[index(i,sign)]   = gamma3qgCMW   ;
	  gamma3gq[index(i,sign)]   = gamma3gqCMW   ;
	  gamma3gg[index(i,sign)]   = gamma3ggCMW   ;
	  
	  gamma4qq[index(i,sign)]   = gamma4qqCMW   ;
	  gamma4qqb[index(i,sign)]  = gamma4qqbCMW  ;
	  gamma4qqp[index(i,sign)]  = gamma4qqpCMW  ;
	  gamma4qqbp[index(i,sign)] = gamma4qqbpCMW ;
	  gamma4qg[index(i,sign)]   = gamma4qgCMW   ;
	  gamma4gq[index(i,sign)]   = gamma4gqCMW   ;
	  gamma4gg[index(i,sign)]   = gamma4ggCMW   ;
	}
    else
      for (int i = 0; i < mellinint::mdim; i++)
	{
	  complex <double> gamma2qqCMW,gamma2qqbCMW,gamma2qqpCMW,gamma2qqbpCMW,gamma2qgCMW,gamma2gqCMW,gamma2ggCMW;
	  complex <double> gamma3qqCMW,gamma3qqbCMW,gamma3qqpCMW,gamma3qqbpCMW,gamma3qgCMW,gamma3gqCMW,gamma3ggCMW;
	  complex <double> gamma4qqCMW,gamma4qqbCMW,gamma4qqpCMW,gamma4qqbpCMW,gamma4qgCMW,gamma4gqCMW,gamma4ggCMW;

	  gamma2qqCMW   = gamma2qq_1[index(i,sign)]   - K1* gamma1qq_1[index(i,sign)]   ;
	  gamma2qqbCMW  = gamma2qqb_1[index(i,sign)]  - K1* gamma1qqb_1[index(i,sign)]  ;
	  gamma2qqpCMW  = gamma2qqp_1[index(i,sign)]  - K1* gamma1qqp_1[index(i,sign)]  ;
	  gamma2qqbpCMW = gamma2qqbp_1[index(i,sign)] - K1* gamma1qqbp_1[index(i,sign)] ;
	  gamma2qgCMW   = gamma2qg_1[index(i,sign)]   - K1* gamma1qg_1[index(i,sign)]   ;
	  gamma2gqCMW   = gamma2gq_1[index(i,sign)]   - K1* gamma1gq_1[index(i,sign)]   ;
	  gamma2ggCMW   = gamma2gg_1[index(i,sign)]   - K1* gamma1gg_1[index(i,sign)]   ;
	  gamma3qqCMW   = gamma3qq_1[index(i,sign)]   - 2.*K1* gamma2qq_1[index(i,sign)]   + 2.*pow(K1,2)*gamma1qq_1[index(i,sign)]   - K2*gamma1qq_1[index(i,sign)]   ;
	  gamma3qqbCMW  = gamma3qqb_1[index(i,sign)]  - 2.*K1* gamma2qqb_1[index(i,sign)]  + 2.*pow(K1,2)*gamma1qqb_1[index(i,sign)]  - K2*gamma1qqb_1[index(i,sign)]  ;
	  gamma3qqpCMW  = gamma3qqp_1[index(i,sign)]  - 2.*K1* gamma2qqp_1[index(i,sign)]  + 2.*pow(K1,2)*gamma1qqp_1[index(i,sign)]  - K2*gamma1qqp_1[index(i,sign)]  ;
	  gamma3qqbpCMW = gamma3qqbp_1[index(i,sign)] - 2.*K1* gamma2qqbp_1[index(i,sign)] + 2.*pow(K1,2)*gamma1qqbp_1[index(i,sign)] - K2*gamma1qqbp_1[index(i,sign)] ;
	  gamma3qgCMW   = gamma3qg_1[index(i,sign)]   - 2.*K1* gamma2qg_1[index(i,sign)]   + 2.*pow(K1,2)*gamma1qg_1[index(i,sign)]   - K2*gamma1qg_1[index(i,sign)]   ;
	  gamma3gqCMW   = gamma3gq_1[index(i,sign)]   - 2.*K1* gamma2gq_1[index(i,sign)]   + 2.*pow(K1,2)*gamma1gq_1[index(i,sign)]   - K2*gamma1gq_1[index(i,sign)]   ;
	  gamma3ggCMW   = gamma3gg_1[index(i,sign)]   - 2.*K1* gamma2gg_1[index(i,sign)]   + 2.*pow(K1,2)*gamma1gg_1[index(i,sign)]   - K2*gamma1gg_1[index(i,sign)]   ;
	  gamma4qqCMW   = gamma4qq_1[index(i,sign)]   - 3.*K1* gamma3qq_1[index(i,sign)]   + (5.*pow(K1,2)-2.*K2)*gamma2qq_1[index(i,sign)]   +(-5.*pow(K1,3)+5.*K1*K2-K3)*gamma1qq_1[index(i,sign)]   ;
	  gamma4qqbCMW  = gamma4qqb_1[index(i,sign)]  - 3.*K1* gamma3qqb_1[index(i,sign)]  + (5.*pow(K1,2)-2.*K2)*gamma2qqb_1[index(i,sign)]  +(-5.*pow(K1,3)+5.*K1*K2-K3)*gamma1qqb_1[index(i,sign)]  ;
	  gamma4qqpCMW  = gamma4qqp_1[index(i,sign)]  - 3.*K1* gamma3qqp_1[index(i,sign)]  + (5.*pow(K1,2)-2.*K2)*gamma2qqp_1[index(i,sign)]  +(-5.*pow(K1,3)+5.*K1*K2-K3)*gamma1qqp_1[index(i,sign)]  ;
	  gamma4qqbpCMW = gamma4qqbp_1[index(i,sign)] - 3.*K1* gamma3qqbp_1[index(i,sign)] + (5.*pow(K1,2)-2.*K2)*gamma2qqbp_1[index(i,sign)] +(-5.*pow(K1,3)+5.*K1*K2-K3)*gamma1qqbp_1[index(i,sign)] ;
	  gamma4qgCMW   = gamma4qg_1[index(i,sign)]   - 3.*K1* gamma3qg_1[index(i,sign)]   + (5.*pow(K1,2)-2.*K2)*gamma2qg_1[index(i,sign)]   +(-5.*pow(K1,3)+5.*K1*K2-K3)*gamma1qg_1[index(i,sign)]   ;
	  gamma4gqCMW   = gamma4gq_1[index(i,sign)]   - 3.*K1* gamma3gq_1[index(i,sign)]   + (5.*pow(K1,2)-2.*K2)*gamma2gq_1[index(i,sign)]   +(-5.*pow(K1,3)+5.*K1*K2-K3)*gamma1gq_1[index(i,sign)]   ;
	  gamma4ggCMW   = gamma4gg_1[index(i,sign)]   - 3.*K1* gamma3gg_1[index(i,sign)]   + (5.*pow(K1,2)-2.*K2)*gamma2gg_1[index(i,sign)]   +(-5.*pow(K1,3)+5.*K1*K2-K3)*gamma1gg_1[index(i,sign)]   ;
	  gamma2qq_1[index(i,sign)]   = gamma2qqCMW   ;
	  gamma2qqb_1[index(i,sign)]  = gamma2qqbCMW  ;
	  gamma2qqp_1[index(i,sign)]  = gamma2qqpCMW  ;
	  gamma2qqbp_1[index(i,sign)] = gamma2qqbpCMW ;
	  gamma2qg_1[index(i,sign)]   = gamma2qgCMW   ;
	  gamma2gq_1[index(i,sign)]   = gamma2gqCMW   ;
	  gamma2gg_1[index(i,sign)]   = gamma2ggCMW   ;
	  gamma3qq_1[index(i,sign)]   = gamma3qqCMW   ;
	  gamma3qqb_1[index(i,sign)]  = gamma3qqbCMW  ;
	  gamma3qqp_1[index(i,sign)]  = gamma3qqpCMW  ;
	  gamma3qqbp_1[index(i,sign)] = gamma3qqbpCMW ;
	  gamma3qg_1[index(i,sign)]   = gamma3qgCMW   ;
	  gamma3gq_1[index(i,sign)]   = gamma3gqCMW   ;
	  gamma3gg_1[index(i,sign)]   = gamma3ggCMW   ;
	  gamma4qq_1[index(i,sign)]   = gamma4qqCMW   ;
	  gamma4qqb_1[index(i,sign)]  = gamma4qqbCMW  ;
	  gamma4qqp_1[index(i,sign)]  = gamma4qqpCMW  ;
	  gamma4qqbp_1[index(i,sign)] = gamma4qqbpCMW ;
	  gamma4qg_1[index(i,sign)]   = gamma4qgCMW   ;
	  gamma4gq_1[index(i,sign)]   = gamma4gqCMW   ;
	  gamma4gg_1[index(i,sign)]   = gamma4ggCMW   ;

	  gamma2qqCMW   = gamma2qq_2[index(i,sign)]   - K1* gamma1qq_2[index(i,sign)]   ;
	  gamma2qqbCMW  = gamma2qqb_2[index(i,sign)]  - K1* gamma1qqb_2[index(i,sign)]  ;
	  gamma2qqpCMW  = gamma2qqp_2[index(i,sign)]  - K1* gamma1qqp_2[index(i,sign)]  ;
	  gamma2qqbpCMW = gamma2qqbp_2[index(i,sign)] - K1* gamma1qqbp_2[index(i,sign)] ;
	  gamma2qgCMW   = gamma2qg_2[index(i,sign)]   - K1* gamma1qg_2[index(i,sign)]   ;
	  gamma2gqCMW   = gamma2gq_2[index(i,sign)]   - K1* gamma1gq_2[index(i,sign)]   ;
	  gamma2ggCMW   = gamma2gg_2[index(i,sign)]   - K1* gamma1gg_2[index(i,sign)]   ;
	  gamma3qqCMW   = gamma3qq_2[index(i,sign)]   - 2.*K1* gamma2qq_2[index(i,sign)]   + 2.*pow(K1,2)*gamma1qq_2[index(i,sign)]   - K2*gamma1qq_2[index(i,sign)]   ;
	  gamma3qqbCMW  = gamma3qqb_2[index(i,sign)]  - 2.*K1* gamma2qqb_2[index(i,sign)]  + 2.*pow(K1,2)*gamma1qqb_2[index(i,sign)]  - K2*gamma1qqb_2[index(i,sign)]  ;
	  gamma3qqpCMW  = gamma3qqp_2[index(i,sign)]  - 2.*K1* gamma2qqp_2[index(i,sign)]  + 2.*pow(K1,2)*gamma1qqp_2[index(i,sign)]  - K2*gamma1qqp_2[index(i,sign)]  ;
	  gamma3qqbpCMW = gamma3qqbp_2[index(i,sign)] - 2.*K1* gamma2qqbp_2[index(i,sign)] + 2.*pow(K1,2)*gamma1qqbp_2[index(i,sign)] - K2*gamma1qqbp_2[index(i,sign)] ;
	  gamma3qgCMW   = gamma3qg_2[index(i,sign)]   - 2.*K1* gamma2qg_2[index(i,sign)]   + 2.*pow(K1,2)*gamma1qg_2[index(i,sign)]   - K2*gamma1qg_2[index(i,sign)]   ;
	  gamma3gqCMW   = gamma3gq_2[index(i,sign)]   - 2.*K1* gamma2gq_2[index(i,sign)]   + 2.*pow(K1,2)*gamma1gq_2[index(i,sign)]   - K2*gamma1gq_2[index(i,sign)]   ;
	  gamma3ggCMW   = gamma3gg_2[index(i,sign)]   - 2.*K1* gamma2gg_2[index(i,sign)]   + 2.*pow(K1,2)*gamma1gg_2[index(i,sign)]   - K2*gamma1gg_2[index(i,sign)]   ;
	  gamma4qqCMW   = gamma4qq_2[index(i,sign)]   - 3.*K1* gamma3qq_2[index(i,sign)]   + (5.*pow(K1,2)-2.*K2)*gamma2qq_2[index(i,sign)]   +(-5.*pow(K1,3)+5.*K1*K2-K3)*gamma1qq_2[index(i,sign)]   ;
	  gamma4qqbCMW  = gamma4qqb_2[index(i,sign)]  - 3.*K1* gamma3qqb_2[index(i,sign)]  + (5.*pow(K1,2)-2.*K2)*gamma2qqb_2[index(i,sign)]  +(-5.*pow(K1,3)+5.*K1*K2-K3)*gamma1qqb_2[index(i,sign)]  ;
	  gamma4qqpCMW  = gamma4qqp_2[index(i,sign)]  - 3.*K1* gamma3qqp_2[index(i,sign)]  + (5.*pow(K1,2)-2.*K2)*gamma2qqp_2[index(i,sign)]  +(-5.*pow(K1,3)+5.*K1*K2-K3)*gamma1qqp_2[index(i,sign)]  ;
	  gamma4qqbpCMW = gamma4qqbp_2[index(i,sign)] - 3.*K1* gamma3qqbp_2[index(i,sign)] + (5.*pow(K1,2)-2.*K2)*gamma2qqbp_2[index(i,sign)] +(-5.*pow(K1,3)+5.*K1*K2-K3)*gamma1qqbp_2[index(i,sign)] ;
	  gamma4qgCMW   = gamma4qg_2[index(i,sign)]   - 3.*K1* gamma3qg_2[index(i,sign)]   + (5.*pow(K1,2)-2.*K2)*gamma2qg_2[index(i,sign)]   +(-5.*pow(K1,3)+5.*K1*K2-K3)*gamma1qg_2[index(i,sign)]   ;
	  gamma4gqCMW   = gamma4gq_2[index(i,sign)]   - 3.*K1* gamma3gq_2[index(i,sign)]   + (5.*pow(K1,2)-2.*K2)*gamma2gq_2[index(i,sign)]   +(-5.*pow(K1,3)+5.*K1*K2-K3)*gamma1gq_2[index(i,sign)]   ;
	  gamma4ggCMW   = gamma4gg_2[index(i,sign)]   - 3.*K1* gamma3gg_2[index(i,sign)]   + (5.*pow(K1,2)-2.*K2)*gamma2gg_2[index(i,sign)]   +(-5.*pow(K1,3)+5.*K1*K2-K3)*gamma1gg_2[index(i,sign)]   ;
	  gamma2qq_2[index(i,sign)]   = gamma2qqCMW   ;
	  gamma2qqb_2[index(i,sign)]  = gamma2qqbCMW  ;
	  gamma2qqp_2[index(i,sign)]  = gamma2qqpCMW  ;
	  gamma2qqbp_2[index(i,sign)] = gamma2qqbpCMW ;
	  gamma2qg_2[index(i,sign)]   = gamma2qgCMW   ;
	  gamma2gq_2[index(i,sign)]   = gamma2gqCMW   ;
	  gamma2gg_2[index(i,sign)]   = gamma2ggCMW   ;
	  gamma3qq_2[index(i,sign)]   = gamma3qqCMW   ;
	  gamma3qqb_2[index(i,sign)]  = gamma3qqbCMW  ;
	  gamma3qqp_2[index(i,sign)]  = gamma3qqpCMW  ;
	  gamma3qqbp_2[index(i,sign)] = gamma3qqbpCMW ;
	  gamma3qg_2[index(i,sign)]   = gamma3qgCMW   ;
	  gamma3gq_2[index(i,sign)]   = gamma3gqCMW   ;
	  gamma3gg_2[index(i,sign)]   = gamma3ggCMW   ;
	  gamma4qq_2[index(i,sign)]   = gamma4qqCMW   ;
	  gamma4qqb_2[index(i,sign)]  = gamma4qqbCMW  ;
	  gamma4qqp_2[index(i,sign)]  = gamma4qqpCMW  ;
	  gamma4qqbp_2[index(i,sign)] = gamma4qqbpCMW ;
	  gamma4qg_2[index(i,sign)]   = gamma4qgCMW   ;
	  gamma4gq_2[index(i,sign)]   = gamma4gqCMW   ;
	  gamma4gg_2[index(i,sign)]   = gamma4ggCMW   ;
	}

  //Compute negative branch
  if (opts.mellin1d)
    for (int m = 0; m < mellinint::mdim; m++)
      {
	int idxp = index(m,mesq::positive);
	int idxm = index(m,mesq::negative);

	gamma1qq[idxm]   = conj(gamma1qq[idxp]);
	gamma1qqqed[idxm]   = conj(gamma1qqqed[idxp]);        // QED
	gamma1qqb[idxm]  = conj(gamma1qqb[idxp]);
	gamma1qqp[idxm]  = conj(gamma1qqp[idxp]);
	gamma1qqbp[idxm] = conj(gamma1qqbp[idxp]);
	gamma1qg[idxm]   = conj(gamma1qg[idxp]);
	gamma1qgamma[idxm]   = conj(gamma1qgamma[idxp]);      // QED
	gamma1gq[idxm]   = conj(gamma1gq[idxp]);
	gamma1gg[idxm]   = conj(gamma1gg[idxp]);

	gamma2qq[idxm]   = conj(gamma2qq[idxp]);
	gamma2qqb[idxm]  = conj(gamma2qqb[idxp]);
	gamma2qqp[idxm]  = conj(gamma2qqp[idxp]);
	gamma2qqbp[idxm] = conj(gamma2qqbp[idxp]);
	gamma2qg[idxm]   = conj(gamma2qg[idxp]);
	gamma2gq[idxm]   = conj(gamma2gq[idxp]);
	gamma2gg[idxm]   = conj(gamma2gg[idxp]);

	gamma3qq[idxm]   = conj(gamma3qq[idxp]);
	gamma3qqb[idxm]  = conj(gamma3qqb[idxp]);
	gamma3qqp[idxm]  = conj(gamma3qqp[idxp]);
	gamma3qqbp[idxm] = conj(gamma3qqbp[idxp]);
	gamma3qg[idxm]   = conj(gamma3qg[idxp]);
	gamma3gq[idxm]   = conj(gamma3gq[idxp]);
	gamma3gg[idxm]   = conj(gamma3gg[idxp]);

	gamma4qq[idxm]   = conj(gamma4qq[idxp]);
	gamma4qqb[idxm]  = conj(gamma4qqb[idxp]);
	gamma4qqp[idxm]  = conj(gamma4qqp[idxp]);
	gamma4qqbp[idxm] = conj(gamma4qqbp[idxp]);
	gamma4qg[idxm]   = conj(gamma4qg[idxp]);
	gamma4gq[idxm]   = conj(gamma4gq[idxp]);
	gamma4gg[idxm]   = conj(gamma4gg[idxp]);
      }
  else
    {
    for (int m = 0; m < mellinint::mdim; m++)
      {
	int idxp = index(m,mesq::positive);
	int idxm = index(m,mesq::negative);

	gamma1qq_1[idxm]   = conj(gamma1qq_1[idxp]);
	gamma1qqqed_1[idxm]   = conj(gamma1qqqed_1[idxp]);     // QED
	gamma1qqb_1[idxm]  = conj(gamma1qqb_1[idxp]);
	gamma1qqp_1[idxm]  = conj(gamma1qqp_1[idxp]);
	gamma1qqbp_1[idxm] = conj(gamma1qqbp_1[idxp]);
	gamma1qg_1[idxm]   = conj(gamma1qg_1[idxp]);
	gamma1qgamma_1[idxm]   = conj(gamma1qgamma_1[idxp]);   // QED
	gamma1gq_1[idxm]   = conj(gamma1gq_1[idxp]);
	gamma1gg_1[idxm]   = conj(gamma1gg_1[idxp]);
	gamma2qq_1[idxm]   = conj(gamma2qq_1[idxp]);
	gamma2qqb_1[idxm]  = conj(gamma2qqb_1[idxp]);
	gamma2qqp_1[idxm]  = conj(gamma2qqp_1[idxp]);
	gamma2qqbp_1[idxm] = conj(gamma2qqbp_1[idxp]);
	gamma2qg_1[idxm]   = conj(gamma2qg_1[idxp]);
	gamma2gq_1[idxm]   = conj(gamma2gq_1[idxp]);
	gamma2gg_1[idxm]   = conj(gamma2gg_1[idxp]);
	gamma3qq_1[idxm]   = conj(gamma3qq_1[idxp]);
	gamma3qqb_1[idxm]  = conj(gamma3qqb_1[idxp]);
	gamma3qqp_1[idxm]  = conj(gamma3qqp_1[idxp]);
	gamma3qqbp_1[idxm] = conj(gamma3qqbp_1[idxp]);
	gamma3qg_1[idxm]   = conj(gamma3qg_1[idxp]);
	gamma3gq_1[idxm]   = conj(gamma3gq_1[idxp]);
	gamma3gg_1[idxm]   = conj(gamma3gg_1[idxp]);
	gamma4qq_1[idxm]   = conj(gamma4qq_1[idxp]);
	gamma4qqb_1[idxm]  = conj(gamma4qqb_1[idxp]);
	gamma4qqp_1[idxm]  = conj(gamma4qqp_1[idxp]);
	gamma4qqbp_1[idxm] = conj(gamma4qqbp_1[idxp]);
	gamma4qg_1[idxm]   = conj(gamma4qg_1[idxp]);
	gamma4gq_1[idxm]   = conj(gamma4gq_1[idxp]);
	gamma4gg_1[idxm]   = conj(gamma4gg_1[idxp]);

	gamma1qq_2[idxm]   = conj(gamma1qq_2[idxp]);
	gamma1qqqed_2[idxm]   = conj(gamma1qqqed_2[idxp]);     // QED
	gamma1qqb_2[idxm]  = conj(gamma1qqb_2[idxp]);
	gamma1qqp_2[idxm]  = conj(gamma1qqp_2[idxp]);
	gamma1qqbp_2[idxm] = conj(gamma1qqbp_2[idxp]);
	gamma1qg_2[idxm]   = conj(gamma1qg_2[idxp]);
	gamma1qgamma_2[idxm]   = conj(gamma1qgamma_2[idxp]);   // QED
	gamma1gq_2[idxm]   = conj(gamma1gq_2[idxp]);
	gamma1gg_2[idxm]   = conj(gamma1gg_2[idxp]);
	gamma2qq_2[idxm]   = conj(gamma2qq_2[idxp]);
	gamma2qqb_2[idxm]  = conj(gamma2qqb_2[idxp]);
	gamma2qqp_2[idxm]  = conj(gamma2qqp_2[idxp]);
	gamma2qqbp_2[idxm] = conj(gamma2qqbp_2[idxp]);
	gamma2qg_2[idxm]   = conj(gamma2qg_2[idxp]);
	gamma2gq_2[idxm]   = conj(gamma2gq_2[idxp]);
	gamma2gg_2[idxm]   = conj(gamma2gg_2[idxp]);
	gamma3qq_2[idxm]   = conj(gamma3qq_2[idxp]);
	gamma3qqb_2[idxm]  = conj(gamma3qqb_2[idxp]);
	gamma3qqp_2[idxm]  = conj(gamma3qqp_2[idxp]);
	gamma3qqbp_2[idxm] = conj(gamma3qqbp_2[idxp]);
	gamma3qg_2[idxm]   = conj(gamma3qg_2[idxp]);
	gamma3gq_2[idxm]   = conj(gamma3gq_2[idxp]);
	gamma3gg_2[idxm]   = conj(gamma3gg_2[idxp]);
	gamma4qq_2[idxm]   = conj(gamma4qq_2[idxp]);
	gamma4qqb_2[idxm]  = conj(gamma4qqb_2[idxp]);
	gamma4qqp_2[idxm]  = conj(gamma4qqp_2[idxp]);
	gamma4qqbp_2[idxm] = conj(gamma4qqbp_2[idxp]);
	gamma4qg_2[idxm]   = conj(gamma4qg_2[idxp]);
	gamma4gq_2[idxm]   = conj(gamma4gq_2[idxp]);
	gamma4gg_2[idxm]   = conj(gamma4gg_2[idxp]);
      }
    }
  
//  for (int sign = mesq::positive; sign <= mesq::negative; sign++)
//    for (int i = 0; i < mellinint::mdim; i++)
//      {
//	cout << endl;
//	cout << i << endl;
//	cout << "gamma 1 " << endl;
//	cout << "qq        " << anomalous::gamma1qq[anomalous::index(i,sign)]/2. << endl;
//	cout << "qg        " << anomalous::gamma1qg[anomalous::index(i,sign)]/2. << endl;
//	cout << "gq        " << anomalous::gamma1gq[anomalous::index(i,sign)]/2. << endl;
//	cout << "gg        " << anomalous::gamma1gg[anomalous::index(i,sign)]/2. << endl;
//	cout << endl;
//	cout << "qq        " << gamma1qq[index(i,sign)] << endl;
//	cout << "qqb       " << gamma1qqb[index(i,sign)] << endl;
//	cout << "qqp       " << gamma1qqp[index(i,sign)] << endl;
//	cout << "qqbp      " << gamma1qqbp[index(i,sign)] << endl;
//	cout << "qg        " << gamma1qg[index(i,sign)] << endl;
//	cout << "gq        " << gamma1gq[index(i,sign)] << endl;
//	cout << "gg        " << gamma1gg[index(i,sign)] << endl;
//	cout << endl;
//
//	cout << "gamma 2 " << endl;
//	cout << "qq        " << (anomalous::gamma2qqV[anomalous::index(i,sign)]+anomalous::gamma2qqS[anomalous::index(i,sign)])/4. << endl;
//	cout << "qqb       " << (anomalous::gamma2qqbV[anomalous::index(i,sign)]+anomalous::gamma2qqbS[anomalous::index(i,sign)])/4. << endl;
//	cout << "qqp       " << anomalous::gamma2qqbS[anomalous::index(i,sign)]/4. << endl;
//	cout << "qqbp      " << anomalous::gamma2qqbS[anomalous::index(i,sign)]/4. << endl;
//	cout << "qg        " << anomalous::gamma2qg[anomalous::index(i,sign)]/4. << endl;
//	cout << "gq        " << anomalous::gamma2gq[anomalous::index(i,sign)]/4. << endl;
//	cout << "gg        " << anomalous::gamma2gg[anomalous::index(i,sign)]/4. << endl;
//	cout << endl;
//	cout << "qq        " << gamma2qq[index(i,sign)] << endl;
//	cout << "qqb       " << gamma2qqb[index(i,sign)] << endl;
//	cout << "qqp       " << gamma2qqp[index(i,sign)] << endl;
//	cout << "qqbp      " << gamma2qqbp[index(i,sign)] << endl;
//	cout << "qg        " << gamma2qg[index(i,sign)] << endl;
//	cout << "gq        " << gamma2gq[index(i,sign)] << endl;
//	cout << "gg        " << gamma2gg[index(i,sign)] << endl;
//	cout << endl;
//
//	cout << "gamma 3 " << endl;
//	cout << "qq        " << gamma3qq[index(i,sign)] << endl;
//	cout << "qqb       " << gamma3qqb[index(i,sign)] << endl;
//	cout << "qqp       " << gamma3qqp[index(i,sign)] << endl;
//	cout << "qqbp      " << gamma3qqbp[index(i,sign)] << endl;
//	cout << "qg        " << gamma3qg[index(i,sign)] << endl;
//	cout << "gq        " << gamma3gq[index(i,sign)] << endl;
//	cout << "gg        " << gamma3gg[index(i,sign)] << endl;
//	cout << endl;
//      }
/*
  for (int sign = mesq::positive; sign <= mesq::negative; sign++)
    for (int i = 0; i < mellinint::mdim; i++)
      {
	cout << endl;
	cout << i << endl;
	cout << "gamma 1 " << endl;
	cout << "qq        " << gamma1qq_1[index(i,sign)] << endl;
	cout << "qqb       " << gamma1qqb_1[index(i,sign)] << endl;
	cout << "qqp       " << gamma1qqp_1[index(i,sign)] << endl;
	cout << "qqbp      " << gamma1qqbp_1[index(i,sign)] << endl;
	cout << "qg        " << gamma1qg_1[index(i,sign)] << endl;
	cout << "gq        " << gamma1gq_1[index(i,sign)] << endl;
	cout << "gg        " << gamma1gg_1[index(i,sign)] << endl;
	cout << endl;

	cout << "gamma 2 " << endl;
	cout << "qq        " << gamma2qq_1[index(i,sign)] << endl;
	cout << "qqb       " << gamma2qqb_1[index(i,sign)] << endl;
	cout << "qqp       " << gamma2qqp_1[index(i,sign)] << endl;
	cout << "qqbp      " << gamma2qqbp_1[index(i,sign)] << endl;
	cout << "qg        " << gamma2qg_1[index(i,sign)] << endl;
	cout << "gq        " << gamma2gq_1[index(i,sign)] << endl;
	cout << "gg        " << gamma2gg_1[index(i,sign)] << endl;
	cout << endl;

	cout << "gamma 3 " << endl;
	cout << "qq        " << gamma3qq_1[index(i,sign)] << endl;
	cout << "qqb       " << gamma3qqb_1[index(i,sign)] << endl;
	cout << "qqp       " << gamma3qqp_1[index(i,sign)] << endl;
	cout << "qqbp      " << gamma3qqbp_1[index(i,sign)] << endl;
	cout << "qg        " << gamma3qg_1[index(i,sign)] << endl;
	cout << "gq        " << gamma3gq_1[index(i,sign)] << endl;
	cout << "gg        " << gamma3gg_1[index(i,sign)] << endl;
	cout << endl;
      }
*/
  
}

void pmom::init()
{
  //if (opts.melup <= 1)
  //  {
  //    allocate();
  //    calc();
  //  }
}

void pmom::release()
{
  //if (opts.melup <= 1)
  //  pmom::free();
}

void pmom::free()
{
  if (opts.mellin1d)
    {
      delete[] gamma1qq;
      delete[] gamma1qqqed;   // QED
      delete[] gamma1qqb;
      delete[] gamma1qqp;
      delete[] gamma1qqbp;
      delete[] gamma1qg;
      delete[] gamma1qgamma;   // QED
      delete[] gamma1gq;
      delete[] gamma1gg;

      delete[] gamma2qq;
      delete[] gamma2qqb;
      delete[] gamma2qqp;
      delete[] gamma2qqbp;
      delete[] gamma2qg;
      delete[] gamma2gq;
      delete[] gamma2gg;

      delete[] gamma3qq;
      delete[] gamma3qqb;
      delete[] gamma3qqp;
      delete[] gamma3qqbp;
      delete[] gamma3qg;
      delete[] gamma3gq;
      delete[] gamma3gg;

      delete[] gamma4qq;
      delete[] gamma4qqb;
      delete[] gamma4qqp;
      delete[] gamma4qqbp;
      delete[] gamma4qg;
      delete[] gamma4gq;
      delete[] gamma4gg;
    }
  else
    {
      delete[] gamma1qq_1;
      delete[] gamma1qqqed_1;  // QED
      delete[] gamma1qqb_1;
      delete[] gamma1qqp_1;
      delete[] gamma1qqbp_1;
      delete[] gamma1qg_1;
      delete[] gamma1qgamma_1; // QED
      delete[] gamma1gq_1;
      delete[] gamma1gg_1;
      delete[] gamma2qq_1;
      delete[] gamma2qqb_1;
      delete[] gamma2qqp_1;
      delete[] gamma2qqbp_1;
      delete[] gamma2qg_1;
      delete[] gamma2gq_1;
      delete[] gamma2gg_1;
      delete[] gamma3qq_1;
      delete[] gamma3qqb_1;
      delete[] gamma3qqp_1;
      delete[] gamma3qqbp_1;
      delete[] gamma3qg_1;
      delete[] gamma3gq_1;
      delete[] gamma3gg_1;
      delete[] gamma4qq_1;
      delete[] gamma4qqb_1;
      delete[] gamma4qqp_1;
      delete[] gamma4qqbp_1;
      delete[] gamma4qg_1;
      delete[] gamma4gq_1;
      delete[] gamma4gg_1;

      delete[] gamma1qq_2;
      delete[] gamma1qqqed_2;  // QED
      delete[] gamma1qqb_2;
      delete[] gamma1qqp_2;
      delete[] gamma1qqbp_2;
      delete[] gamma1qg_2;
      delete[] gamma1qgamma_2; // QED
      delete[] gamma1gq_2;
      delete[] gamma1gg_2;
      delete[] gamma2qq_2;
      delete[] gamma2qqb_2;
      delete[] gamma2qqp_2;
      delete[] gamma2qqbp_2;
      delete[] gamma2qg_2;
      delete[] gamma2gq_2;
      delete[] gamma2gg_2;
      delete[] gamma3qq_2;
      delete[] gamma3qqb_2;
      delete[] gamma3qqp_2;
      delete[] gamma3qqbp_2;
      delete[] gamma3qg_2;
      delete[] gamma3gq_2;
      delete[] gamma3gg_2;
      delete[] gamma4qq_2;
      delete[] gamma4qqb_2;
      delete[] gamma4qqp_2;
      delete[] gamma4qqbp_2;
      delete[] gamma4qg_2;
      delete[] gamma4gq_2;
      delete[] gamma4gg_2;
    }
}
